----------------------------------------------------------------------
--                                                                  --
-- Copyright(c) Hynix Semiconductor Inc. 2002. All rights reserved. --
--                                                                  --
--        VHDL Simulation Model of 4nd Gen. 128M SDR Device         --
--                    (4 Banks X 4M X 8 I/Os)                       --
--                                                                  --
--                   PART : HY57V28820HCT-6/K/H/8/P/S              --
--                                                                  --
--                     DATE : 2002. 11. 30                          --
--                                                                  --
--          Developer : Jae-Youl Lee, Memory R&D Division           --
--                                                                  --
----------------------------------------------------------------------
---------------------------------------
LIBRARY IEEE;
USE STD.textio.all;
USE IEEE.STD_LOGIC_1164.all;
USE IEEE.STD_LOGIC_MISC.all;
USE IEEE.STD_LOGIC_TEXTIO.all;
USE IEEE.STD_LOGIC_ARITH.all;
USE IEEE.STD_LOGIC_UNSIGNED.all;
USE WORK.HY57V28820HCT_PACK.all; --?
----------------------------------------------------------------------------------
---------------------------------------
Entity HY57V28820HCT is --? device name
  Generic (
    Part_Number : PART_NUM_TYPE := S);
  Port (  VDD1   :  in      std_logic;
          DQ0    :  inout   std_logic := 'Z';
          VDDQ3  :  in      std_logic;
          NC4    :  in      std_logic;
          DQ1    :  inout   std_logic := 'Z';
          VSSQ6  :  in      std_logic;
          NC7    :  in      std_logic;
          DQ2    :  inout   std_logic := 'Z';
          VDDQ9  :  in      std_logic;
          NC10   :  in      std_logic;
          DQ3    :  inout   std_logic := 'Z';
          VSSQ12 :  in      std_logic;
          NC13   :  in      std_logic;
          VDD14  :  in      std_logic;
          NC15   :  in      std_logic;
          WEB    :  in      std_logic;
          CASB   :  in      std_logic;
          RASB   :  in      std_logic;
          CSB    :  in      std_logic;
          BA0    :  in      std_logic;
          BA1    :  in      std_logic;
          A10    :  in      std_logic;
          A0     :  in      std_logic;
          A1     :  in      std_logic;
          A2     :  in      std_logic;
          A3     :  in      std_logic;
          VDD27  :  in      std_logic;
          VSS28  :  in      std_logic;
          A4     :  in      std_logic;
          A5     :  in      std_logic;
          A6     :  in      std_logic;
          A7     :  in      std_logic;
          A8     :  in      std_logic;
          A9     :  in      std_logic;
          A11    :  in      std_logic;
          NC36   :  in      std_logic;
          CKE    :  in      std_logic;
          CLK    :  in      std_logic;
          DQM    :  in      std_logic;
          NC40   :  in      std_logic;
          VSS41  :  in      std_logic;
          NC42   :  in      std_logic;
          VDDQ43 :  in      std_logic;
          DQ4    :  inout   std_logic := 'Z';
          NC45   :  in      std_logic;
          VSSQ46 :  in      std_logic;
          DQ5    :  inout   std_logic := 'Z';
          NC48   :  in      std_logic;
          VDDQ49 :  in      std_logic;
          DQ6    :  inout   std_logic := 'Z';
          NC51   :  in      std_logic;
          VSSQ52 :  in      std_logic;
          DQ7    :  inout   std_logic := 'Z';
          VSS54  :  in      std_logic);
End HY57V28820HCT;
----------------------------------------------------------------------------------
---------------------------------------
Architecture Behavioral_Model of HY57V28820HCT is

signal ReadFix : integer := 4;
signal real_CA1 : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0);
signal M1 : integer;
signal Din1 : std_logic_vector((WORD_SIZE-1) DOWNTO 0);
signal DATA1 : std_logic_vector((WORD_SIZE-1) DOWNTO 0);
signal WRITEDRIVEARRAR_CHECK0 : std_logic_vector((WORD_SIZE-1) DOWNTO 0);
signal WRITEDRIVEARRAR_CHECK1 : std_logic_vector((WORD_SIZE-1) DOWNTO 0);
signal WRITEDRIVEARRAR_CHECK2 : std_logic_vector((WORD_SIZE-1) DOWNTO 0);
signal CLKY : std_logic;
signal cur_time : time := 0 ns;
signal bstop_write : std_logic := '0';
signal Ref_time, clk_cycle_rising : time := 0 ns;
signal tmp_act_trans0, tmp_act_trans1, tmp_act_trans2, tmp_act_trans3, RefCnt_trans : std_logic := '0';
signal PCG_WriteFin, PCG_ReadFin1, PCG_ReadFin2 : std_logic := '0';
signal CLK_INT_DLY : std_logic := '0';
signal tmp_ref_addr1 : std_logic_vector((NUM_OF_ROW_ADD - 1) downto 0) := (others => '0');
signal tmp_ref_addr2 : std_logic_vector((NUM_OF_ROW_ADD + NUM_OF_BANK_ADD - 1) downto 0) := (others => '0');
signal RefChkTimeInit : boolean := FALSE;
signal refresh_check : REF_CHECK;
signal real_col_addr : std_logic_vector((NUM_OF_COL_ADD - 1) downto 0) := (others => '0');
signal CKE_last_zero : std_logic := '0';
signal CLK_INT : std_logic := '0';
signal Read_CA, Write_CA : std_logic := '0';
signal d_mask1, d_mask2 : std_logic := '0';
signal DQtransaction, tmp_w_trans0, tmp_w_trans1, tmp_w_trans2, tmp_w_trans3 : std_logic := '0';
signal RA_Activated_B0 : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => 'U');
signal RA_Activated_B1 : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => 'U');
signal RA_Activated_B2 : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => 'U');
signal RA_Activated_B3 : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => 'U');
signal APBA, EXEBA, IMSIBA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'U'); --?
signal CKEN : CKE_TYPE := (others => '1');
signal SA_ARRAY : SA_ARRAY_TYPE;
signal SA_ARRAY_A0 : SA_TYPE;
signal SA_ARRAY_A1 : SA_TYPE;
signal SA_ARRAY_A2 : SA_TYPE;
signal SA_ARRAY_A3 : SA_TYPE;
signal SA_ARRAY_W0 : SA_TYPE;
signal SA_ARRAY_W1 : SA_TYPE;
signal SA_ARRAY_W2 : SA_TYPE;
signal SA_ARRAY_W3 : SA_TYPE;
signal WD_ARRAY : SA_ARRAY_TYPE;
signal RefCnt : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => '0');
signal Clk_Cycle_Chk : boolean := FALSE;
signal ModeRegisterSetFlag : boolean := FALSE;
signal ModeRegisterFlag : boolean := FALSE;
signal BankActivateFlag : boolean := FALSE;
signal BankActivateFinFlag : boolean := FALSE;
signal BankActivatedFlag : std_logic_vector ((NUM_OF_BANKS - 1) downto 0) := (others => '0');
signal ReadFlag : boolean := FALSE;
signal WriteFlag : boolean := FALSE;
signal DataQueue : PIPE_TYPE;
signal PrechargeFlag : boolean := FALSE;
signal AutoPrechargeFlag : boolean := FALSE;
signal PrechargeFinFlag : boolean := FALSE;
signal PrechargeAllFlag : boolean := FALSE;
signal PrechargeAllFinFlag : boolean := FALSE;
signal ReadFinFlag : boolean := FALSE;
signal WriteFinFlag : boolean := FALSE;
signal BurstStopFlag : boolean := FALSE;
signal AutoRefFlag : boolean := FALSE;
signal SelfRefFlag : boolean := FALSE;
signal SelfRefExtFlag : boolean := FALSE;
signal PUSCheckFinFlag : boolean := FALSE;
signal PcgPwrDnFlag : boolean := FALSE;
signal PcgPwrDnExtFlag : boolean := FALSE;
signal ActPwrDnFlag : boolean := FALSE;
signal ActPwrDnExtFlag : boolean := FALSE;
signal CurrentReadBurstIndex : integer := 0;
signal CurrentWriteBurstIndex : integer := 0;
signal CurrentState : STATE_TYPE := IDLE;
signal MROpCode : MROPCODE_TYPE := (others => 'U');
signal ModeRegister, PrevModeRegister : MODE_REGISTER := (          
  WRITE_MODE => BRBW, --? maximum CL value
  CAS_LATENCY => 3,                               
  BURST_MODE => SEQUENTIAL,                     
  BURST_LENGTH => 4                            
  );                                          
signal CLKStableFlag : boolean := FALSE;
signal last_aref_time : time := 0 ns;
signal clk_cycle : time := 0 ns;
signal clk_last_rising : time := 0 ns;
signal clk_last_falling : time := 0 ns;
signal b0_last_activate : time := 0 ns;
signal b1_last_activate : time := 0 ns;
signal b2_last_activate : time := 0 ns;
signal b3_last_activate : time := 0 ns;
signal b0_last_precharge : time := 0 ns;
signal b1_last_precharge : time := 0 ns;
signal b2_last_precharge : time := 0 ns;
signal b3_last_precharge : time := 0 ns;
signal b0_last_column_access : time := 0 ns;
signal b1_last_column_access : time := 0 ns;
signal b2_last_column_access : time := 0 ns;
signal b3_last_column_access : time := 0 ns;
signal b0_last_data_in : time := 0 ns;
signal b1_last_data_in : time := 0 ns;
signal b2_last_data_in : time := 0 ns;
signal b3_last_data_in : time := 0 ns;
signal last_mrs_set : time := 0 ns;
signal last_data_in : time := 0 ns;
signal last_pd_ext : time := 0 ns;
signal last_sr_ext : time := 0 ns;
signal tCCD : time := 0 ns;
signal tDPL : time := 0 ns;
signal tDAL : time := 0 ns;
signal tMRD : time := 0 ns;
signal tPDE : time := 0 ns;
signal tSRE : time := 0 ns;
signal cke_ch : time := 0 ns;
signal rasb_ch : time := 0 ns;
signal casb_ch : time := 0 ns;
signal web_ch : time := 0 ns;
signal csb_ch : time := 0 ns;
signal dqm_ch : time := 0 ns;
signal a0_ch : time := 0 ns;
signal a1_ch : time := 0 ns;
signal a2_ch : time := 0 ns;
signal a3_ch : time := 0 ns;
signal a4_ch : time := 0 ns;
signal a5_ch : time := 0 ns;
signal a6_ch : time := 0 ns;
signal a7_ch : time := 0 ns;
signal a8_ch : time := 0 ns;
signal a9_ch : time := 0 ns;
signal a10_ch : time := 0 ns;
signal a11_ch : time := 0 ns;
signal ba0_ch : time := 0 ns;
signal ba1_ch : time := 0 ns;
signal dq0_ch : time := 0 ns;
signal dq1_ch : time := 0 ns;
signal dq2_ch : time := 0 ns;
signal dq3_ch : time := 0 ns;
signal dq4_ch : time := 0 ns;
signal dq5_ch : time := 0 ns;
signal dq6_ch : time := 0 ns;
signal dq7_ch : time := 0 ns;

Begin

----------------------------------------------------------------------------------
------------------------------------
CLK_CYCLE_CHECK : process(CLK)----------------------------------------------------------------------
---CLK_CYCLE_CHECK

begin
  if (Clk_Cycle_Chk = FALSE) then
    if (CLK'EVENT and CLK = '1') then
      if (clk_cycle_rising = 0 ns) then
        clk_cycle_rising <= now;
      else
        clk_cycle <= now - clk_cycle_rising;
        Clk_Cycle_Chk <= TRUE;
      end if;
    end if;
  end if;
end 
Process;--------------------------------------------------------------------------
-----------------CLK_CYCLE_CHECK INT_CLK_GENERATION : 
process(CLK)
-------------------------------------------------------------------INT_CLK_GENERATION

begin
  if (CLK'EVENT and CLK = '1' and CLK'LAST_VALUE = '0') then
    if (CKE = '1' and CKE_last_zero = '0') then
      CLK_INT <= '1';
    elsif (CKE = '1' and CKE_last_zero = '1') then
      CLK_INT <= '0';
      CKE_last_zero <= '0';
    elsif (CKE = '0' and CKE_last_zero = '0') then
      CLK_INT <= '1';
      CKE_last_zero <= '1';
    elsif (CKE = '0' and CKE_last_zero = '1') then
      CLK_INT <= '0';
      CKE_last_zero <= '1'; 
    end if;
  elsif (CLK'EVENT and CLK = '0' and CLK'LAST_VALUE = '1') then
    CLK_INT <= '0';
  end if;
end 
Process;
--------------------------------------------------------------------------
-------------INT_CLK_GENERATION    

CLK_INT_DLY <= CLK_INT after 2 ns;

REFRESH_TIME_CHECK : process(tmp_ref_addr1, tmp_ref_addr2)
-----------------------------------------REFRESH_TIME_CHECK

variable i, j : integer := 0;

begin
  if (RefChkTimeInit = FALSE) then
    loop
      exit when (i > (NUM_OF_BANKS - 1));
      j := 0;
      loop
        exit when (j >= NUM_OF_ROWS);
        refresh_check (i, j) <= 0 ns;
        j := j + 1;
      end loop;
      i := i + 1;
    end loop;
    RefChkTimeInit <= TRUE;
  end if;
  i := 0;
  j := 0;
  if (tmp_ref_addr1'event) then
    refresh_check (0, conv_integer(tmp_ref_addr1((NUM_OF_ROW_ADD - 1) downto 0))) <= now;
    refresh_check (1, conv_integer(tmp_ref_addr1((NUM_OF_ROW_ADD - 1) downto 0))) <= now;
    refresh_check (2, conv_integer(tmp_ref_addr1((NUM_OF_ROW_ADD - 1) downto 0))) <= now;
    refresh_check (3, conv_integer(tmp_ref_addr1((NUM_OF_ROW_ADD - 1) downto 0))) <= now;
  elsif (tmp_ref_addr2'event) then
    refresh_check (conv_integer(tmp_ref_addr2(1 downto 0)), 
                   conv_integer(tmp_ref_addr2((NUM_OF_ROW_ADD + NUM_OF_BANK_ADD - 1) downto 2))) <= now;
  end if;
end 
process;--------------------------------------------------------------------------
-------------REFRESH_TIME_CHECK


--------------------------------------------------------------------------------CKE(-1) AND CKE(0) SIGNAL EVALUATION
CKE_EVAL : process (CLK)

begin
  if (CLK'EVENT and CLK = '0' and CLK'LAST_VALUE = '1') then
    CKEN(-1) <= CKEN(0);
  elsif (CLK'EVENT and CLK = '1' and CLK'LAST_VALUE = '0') then
    CKEN(0) <= CKE;
  end if;
end 
process;--------------------------------------------------------------------CKE(-1) AND CKE(0) SIGNAL EVALUATION

----------------------------------------------------------------------------------
---------------------STATE_MACHINE
STATE_MACHINE : process (PCG_WriteFin, WriteFinFlag, CLK, CLK_INT, CKE, 
BankActivateFinFlag, PrechargeFinFlag, PrechargeAllFinFlag, SelfRefExtFlag, 
BankActivatedFlag, ReadFinFlag, bstop_write)

variable ApCount : integer := 0;
variable ChipSelectBar : std_logic := '0';
variable RowAddrStrobeBar : std_logic := '0';
variable ColAddrStrobeBar : std_logic := '0';
variable WriteEnableBar : std_logic := '0';
variable Address10 : std_logic := '0';
variable ClockEnable : CKE_TYPE := (others => '0');
variable NextState : STATE_TYPE := IDLE;
variable CurrentCommand : COMMAND_TYPE := NOP;
variable OpCode : MROPCODE_TYPE := (others => 'X');
variable MR : MODE_REGISTER;
variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable BankActFlag : std_logic_vector((NUM_OF_BANKS - 1) downto 0) := (others => '0');

begin
  if (CLK = '1' and CLK'EVENT and CLK'LAST_VALUE = '0') then
    BA := (BA1, BA0);
    ClockEnable(-1) := CKEN(-1);
    ClockEnable(0) := CKE;
    if ClockEnable (-1) = '1' then
      ChipSelectBar := CSB;
      RowAddrStrobeBar := RASB;
      ColAddrStrobeBar := CASB;
      WriteEnableBar := WEB;
      Address10 := A10;  --?
      BankActFlag := BankActivatedFlag;
    end if;
    COMMAND_DECODE (ChipSelectBar, RowAddrStrobeBar, ColAddrStrobeBar,
      WriteEnableBar, Address10, ClockEnable, CurrentCommand, BankActFlag);
    case CurrentState is
      when IDLE =>
        case CurrentCommand is
          when MRS =>
            assert (PUSCheckFinFlag = TRUE) report
            "ERROR : (STATE_MACHINE) : Initialization with Power Up Sequence Error!"
            severity error;
            ModeRegisterSetFlag <= FALSE;
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0); --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
            ModeRegisterSetFlag <= TRUE;
            NextState := IDLE;
          when PD =>
            PcgPwrDnFlag <= TRUE;
            NextState := PWRDN;
          when SREF =>
            SelfRefFlag <= TRUE;
            NextState := SLFREF;
          when ACT =>
            BankActivateFlag <= TRUE;
            NextState := RACT;
          when AREF =>
            AutoRefFlag <= TRUE, FALSE after 2 ns;
            NextState := IDLE;
          when NOP =>
            NextState := IDLE;  
          when PCGA =>
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when PCG =>
            PrechargeFlag <= TRUE;
            NextState := IDLE;
          when CSPND =>
            NextState := CLKSPND;
          when RD|RDA =>
            Read_CA <= '1';
            ReadFlag <= TRUE;
            NextState := READ;
          when WR|WRA =>
            Write_CA <= '1';
            WriteFlag <= TRUE;
            NextState := WRITE;
          when others =>
            assert false
            report "NOTE : (1STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
        end case;
      when READA =>
        ApCount := ApCount + 1;
        case CurrentCommand is
          when MRS =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := READA;
            end if;
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
          when RD =>
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            NextState := READ;
          when WR =>
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
-- EDIT
--          WriteFlag <= TRUE after 2.5 ns;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            NextState := WRITE;
          when RDA =>                            
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := READA;
          when WRA =>                          
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
-- EDIT
--          WriteFlag <= TRUE after 2.5 ns;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := WRITEA;
          when PCG =>
            if (EXEBA = (BA1, BA0)) then
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle, 
                                             FALSE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle + 2 ns;
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
                NextState := IDLE;
              else
                NextState := READA;
              end if;
            end if;
            PrechargeFlag <= TRUE;
          when PCGA =>
            APBA <= EXEBA;
            AutoPrechargeFlag <= transport TRUE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle, 
                                           FALSE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle + 2 ns;
            ApCount := 0;
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when ACT =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              NextState := RACT;
            else
              NextState := READA;
            end if;
            BankActivateFlag <= TRUE;
          when BSTOP =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
            else
              AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns; 
              BurstStopFlag <= transport TRUE after (ModeRegister.CAS_LATENCY - 1)*clk_cycle,
                                         FALSE after (ModeRegister.CAS_LATENCY)*clk_cycle;
            end if;
            if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
              NextState := IDLE;
            else
              NextState := RACT;
            end if;
            ApCount := 0;
          when CSPND =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
            end if;
            NextState := RDASPND;
          when NOP =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := READA;
            end if;
          when CKEEXIT =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := READA;
            end if;
          when others =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := READA;
            end if;
            assert false report
            "NOTE : (2STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
        end case;
      when RDASPND =>
        if (CurrentCommand = CKEEXIT) then
          if (BankActivatedFlag(conv_integer(EXEBA)) = '1') then
            NextState := READA;
          else
            if (BankActivatedFlag = "0000") then
              NextState := IDLE;
            else
              NextState := RACT;
            end if;
          end if;
        elsif (CurrentCommand = NOP) then
          NextState := RDASPND;
        else
          assert false report
          "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := RDASPND;
        end if;
      when WRITEA =>
        ApCount := ApCount + 1;
        case CurrentCommand is
          when MRS =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := WRITEA;
            end if;
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
          when RD =>
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            NextState := READ;
          when WR =>
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            WriteFlag <= TRUE;
            Write_CA <= '1'; 
            NextState := WRITE;
          when RDA =>                            
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            ReadFlag <= TRUE;
            Read_CA <= '1'; 
            EXEBA <= (BA1, BA0);
            NextState := READA;
          when WRA =>                          
            if (EXEBA = (BA1, BA0)) then
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
              else
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns;
                ApCount := 0;
              end if;
            end if;
            WriteFlag <= TRUE;
            Write_CA <= '1'; 
            EXEBA <= (BA1, BA0);
            NextState := WRITEA;
          when PCG =>
            if (EXEBA = (BA1, BA0)) then
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle, 
                                             FALSE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle + 2 ns;
              ApCount := 0;
            else
              if (ApCount = PrevModeRegister.BURST_LENGTH) then
                APBA <= EXEBA;
                AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
                ApCount := 0;
                NextState := IDLE;
              else
                NextState := WRITEA;
              end if;
            end if;
            PrechargeFlag <= TRUE;
          when PCGA =>
            APBA <= EXEBA;
            AutoPrechargeFlag <= transport TRUE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle, 
                                           FALSE after (PrevModeRegister.BURST_LENGTH - ApCount)*clk_cycle + 2 ns;
            ApCount := 0;
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when ACT =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              NextState := RACT;
            else
              NextState := WRITEA;
            end if;
            BankActivateFlag <= TRUE;
          when BSTOP =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
            else
              AutoPrechargeFlag <= transport TRUE after clk_cycle, FALSE after clk_cycle + 2 ns; 
              BurstStopFlag <= transport TRUE, FALSE after clk_cycle;
            end if;
            if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
              NextState := IDLE;
            else
              NextState := RACT;
            end if;
            ApCount := 0; 
          when CSPND =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
            end if;
            NextState := WRASPND;
          when NOP =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := WRITEA;
            end if;
          when CKEEXIT =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := WRITEA;
            end if;
          when others =>
            if (ApCount = PrevModeRegister.BURST_LENGTH) then
              APBA <= EXEBA;
              AutoPrechargeFlag <= transport TRUE, FALSE after 2 ns;
              ApCount := 0;
              if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                  (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            else
              NextState := WRITEA;
            end if;
            assert false report
            "NOTE : (3STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
        end case;
      when WRASPND =>
        if (CurrentCommand = CKEEXIT) then
          if (BankActivatedFlag(conv_integer(EXEBA)) = '1') then
            NextState := WRITEA;
          else
            if (BankActivatedFlag = "0000") then
              NextState := IDLE;
            else
              NextState := RACT;
            end if;
          end if;
        elsif (CurrentCommand = NOP) then
          NextState := WRASPND;
        else
          assert false report
          "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := WRASPND;
        end if;
      when CLKSPND =>
        if (CurrentCommand = CKEEXIT) then
          NextState := IDLE;
        elsif (CurrentCommand = NOP) then
          NextState := CLKSPND;
        else
          assert False
          report "NOTE : (4STATE_MACHINE) : Illegal Command Issued. Command Ignored."
          severity note;
        end if;
      when SLFREF =>
        if (CurrentCommand = CKEEXIT) then
          SelfRefExtFlag <= transport TRUE, FALSE after clk_cycle + 2 ns;
          last_sr_ext <= transport now after 1 ns;
          NextState := IDLE;
        elsif (CurrentCommand = NOP) then
          SelfRefFlag <= TRUE;
          NextState := SLFREF;
        else
          assert false
          report "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := SLFREF;
        end if;
      when PWRDN =>
        if (CurrentCommand = CKEEXIT) then
          PcgPwrDnExtFlag <= TRUE;
          last_pd_ext <= transport now after 1 ns;
          NextState := IDLE;
        elsif (CurrentCommand = NOP) then
          PcgPwrDnFlag <= TRUE;
          NextState := PWRDN;
        else
          assert false
          report "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := SLFREF;
        end if;
      when RACT =>
        case CurrentCommand is
          when MRS =>
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
          when CSPND =>
            ActPwrDnFlag <= TRUE;
            NextState := ACTSPND;
          when ACT =>
            BankActivateFlag <= TRUE;
            NextState := RACT;
          when RD =>
            ReadFlag <= TRUE;
            Read_CA <= '1';
            NextState := READ;
          when RDA =>
            ApCount := 0;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := READA;
          when WR =>
            WriteFlag <= TRUE;
            Write_CA <= '1';
            NextState := WRITE;
          when WRA =>
            ApCount := 0;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := WRITEA;
          when PCG =>
            PrechargeFlag <= TRUE;
            if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
                (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
              NextState := IDLE;
            else
              NextState := RACT;
            end if;
          when PCGA =>
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when NOP =>
            NextState := RACT;
          when AREF =>
            NextState := RACT;
          when others =>
            assert false
            report "NOTE : (5STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
          end case;
      when ACTSPND =>
        if (CurrentCommand = CKEEXIT) then
          ActPwrDnExtFlag <= transport TRUE, FALSE after 2 ns;
          NextState := RACT;
        elsif (CurrentCommand = NOP) then
          ActPwrDnFlag <= transport TRUE, FALSE after 2 ns;
          NextState := ACTSPND;
        else
          assert false
          report "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := ACTSPND;
        end if;
      when READ =>
        case CurrentCommand is
          when MRS =>
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
          when RD =>
            ReadFlag <= TRUE; -- after 1 ns; -- trying to fix it here
            Read_CA <= '1';
            NextState := READ;
          when RDA =>
            ApCount := 0;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := READA;
          when WR =>
-- EDIT
--          WriteFlag <= TRUE after 2.5 ns;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            NextState := WRITE;
          when WRA =>
            ApCount := 0;
-- EDIT
--          WriteFlag <= TRUE after 2.5 ns;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := WRITEA;
          when ACT =>
            BankActivateFlag <= TRUE;
            NextState := READ;
          when PCG =>
-- EDIT
--          PrechargeFlag <= TRUE;
--          if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
--              (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
-- EDIT
            PrechargeFlag <= TRUE;
            IMSIBA <= (BA1, BA0);
            if (BankActivatedFlag(conv_integer(IMSIBA)) = '1') then
              if (((BankActivatedFlag = "0001") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "0010") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "0100") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "1000") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            end if;
          when PCGA =>
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when BSTOP =>
            BurstStopFlag <= transport TRUE after (ModeRegister.CAS_LATENCY - 1)*clk_cycle,
                                       FALSE after (ModeRegister.CAS_LATENCY)*clk_cycle;
-- EDIT
--          if (WriteFlag = FALSE) then
              NextState := RACT;
--          end if;
          when CSPND =>
            NextState := RDSPND;
          when NOP =>
            NextState := READ;
          when CKEEXIT =>
            NextState := READ;
          when others =>
            assert false report
            "NOTE : (6STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
        end case;
        if ((CurrentCommand /= RD) and (NextState = READ) and (ReadFinFlag = TRUE)) then
          NextState := RACT;
        end if;
      when RDSPND =>
        if (CurrentCommand = CKEEXIT) then
          NextState := READ;
        elsif (CurrentCommand = NOP) then
          NextState := RDSPND;
        else
          assert false report
          "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := RDSPND;
        end if;
      when WRITE =>
        case CurrentCommand is
          when MRS =>
            last_mrs_set <= transport now after 1 ns;
            OpCode := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
            MODE_REGISTER_SET (OpCode, MR);
            MROpCode <= OpCode;
            ModeRegister <= MR;
          when RD =>
            ReadFlag <= TRUE;
            Read_CA <= '1';
            NextState := READ;
          when RDA =>
            ApCount := 0;
            ReadFlag <= TRUE;
            Read_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := READA;
          when WR =>
            WriteFlag <= TRUE;
            Write_CA <= '1';
            NextState := WRITE;
          when WRA =>
            ApCount := 0;
            WriteFlag <= TRUE;
            Write_CA <= '1';
            EXEBA <= (BA1, BA0);
            NextState := WRITEA;
          when ACT =>
            BankActivateFlag <= TRUE;
            NextState := WRITE;
          when PCG =>
-- EDIT
--          PrechargeFlag <= TRUE;
--          if ((BankActivatedFlag = "0001") or (BankActivatedFlag = "0010") or
--              (BankActivatedFlag = "0100") or (BankActivatedFlag = "1000")) then
-- EDIT
            PrechargeFlag <= TRUE;
            IMSIBA <= (BA1, BA0);
            if (BankActivatedFlag(conv_integer(IMSIBA)) = '1') then
              if (((BankActivatedFlag = "0001") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "0010") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "0100") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))
                  or ((BankActivatedFlag = "1000") and (BankActivatedFlag(conv_integer(IMSIBA)) = '1'))) then
                NextState := IDLE;
              else
                NextState := RACT;
              end if;
            end if;
          when PCGA =>
            PrechargeAllFlag <= TRUE;
            NextState := IDLE;
          when BSTOP =>
            BurstStopFlag <= transport TRUE, FALSE after clk_cycle;
-- EDIT
--          if (ReadFlag = FALSE) then
              NextState := RACT;
--          end if;
          when CSPND =>
            NextState := WRSPND;
          when NOP =>
            NextState := WRITE;
          when CKEEXIT =>
            NextState := WRITE;
          when others =>
            assert false report
            "NOTE : (7STATE_MACHINE) : Illegal Command Issued. Command Ignored."
            severity note;
        end case;
        if ((CurrentCommand /= WR) and (NextState = WRITE) and (WriteFinFlag = TRUE)) then
          NextState := RACT;
        end if;
      when WRSPND =>
        if (CurrentCommand = CKEEXIT) then
          NextState := WRITE;
        elsif (CurrentCommand = NOP) then
          NextState := WRSPND;
        else
          assert false
          report "NOTE : (STATE_MACHINE) : Command ignored"
          severity note;
          NextState := WRSPND;
        end if;
      when others =>
        assert false report
        "ERROR : (STATE_MACHINE) : Impossible State" severity error;
        NextState := ERROR;
      end case;
  else
  end if;
  if (CLK_INT'EVENT and CLK_INT = '0') then
    if (Read_CA = '1') then
      Read_CA <= '0';
    end if;
    if (Write_CA = '1') then
      Write_CA <= '0';
    end if;
  end if;
  if ((bstop_write'EVENT and bstop_write = '1') or (WriteFinFlag'EVENT and WriteFinFlag = TRUE) or
      (PCG_WriteFin'EVENT and PCG_WriteFin = '1')) then
    WriteFlag <= FALSE;
  end if;
  if (BankActivateFinFlag = TRUE) then
    BankActivateFlag <= FALSE after 2 ns;
  end if;
  if (PrechargeFinFlag = TRUE) then
    PrechargeFlag <= FALSE after 5 ns;
  end if;
  if (ReadFinFlag'EVENT and ReadFinFlag = TRUE) then
    ReadFlag <= FALSE;
    if (CurrentState /= READA) then
      if (BankActivatedFlag = "0000") then
        NextState := IDLE;
-- EDIT
--    else
--      NextState := RACT;
      elsif (WriteFlag = FALSE) then
        NextState := RACT;
      end if;
    end if;
  end if;
  if (PrechargeAllFinFlag = TRUE) then
    PrechargeAllFlag <= FALSE after 5 ns;
  end if;
  if (SelfRefExtFlag = TRUE) then
    SelfRefFlag <= FALSE after 2 ns;
  end if;
  if (BankActivatedFlag'EVENT) then
    if (BankActivatedFlag = "0000") then
      NextState := IDLE;
    end if;
  end if;
  CurrentState <= NextState;
end process; 
----------------------------------------------------------------------------------
--------STATE_MACHINE

----------------------------------------------------------------------------------
------------MEMORY_BANK_ACTIVATE_PRECHARGE
MEMORY_BANK_ACTIVATE_PRECHARGE : process (BankActivateFlag, CLK_INT, PrechargeFlag, AutoPrechargeFlag, PrechargeAllFlag)

variable BA : std_logic_vector ((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable RA : std_logic_vector ((NUM_OF_ROW_ADD - 1) downto 0) := (others => 'X');
variable SenseAmpArray : SA_ARRAY_TYPE;
variable DATA : std_logic_vector ((WORD_SIZE - 1) downto 0) := (others => 'X');
variable SA : SA_TYPE;
variable TMP_BUF : std_logic_vector ((WORD_SIZE - 1) downto 0) := (others => '0');
variable i, j, l, u : integer := 0;
variable MEM_CELL_ARRAY0, MEM_CELL_ARRAY1 : MEM_CELL_TYPE;
variable MEM_CELL_ARRAY2, MEM_CELL_ARRAY3 : MEM_CELL_TYPE;

begin
  if (BankActivateFlag'EVENT and BankActivateFlag = TRUE) Then
    BA := (BA1, BA0); 
    BankActivatedFlag(conv_integer(BA)) <= '1';
  End if;
  if (CLK_INT'EVENT and CLK_INT = '0') then
    if ((BankActivateFlag = TRUE) or ((BankActivateFlag = FALSE) and (tmp_act_trans0 = '1'))) Then
      tmp_act_trans0 <= '0';
    end if;
    if ((BankActivateFlag = TRUE) or ((BankActivateFlag = FALSE) and (tmp_act_trans1 = '1'))) Then
      tmp_act_trans1 <= '0';
    end if;
    if ((BankActivateFlag = TRUE) or ((BankActivateFlag = FALSE) and (tmp_act_trans2 = '1'))) Then
      tmp_act_trans2 <= '0';
    end if;
    if ((BankActivateFlag = TRUE) or ((BankActivateFlag = FALSE) and (tmp_act_trans3 = '1'))) Then
      tmp_act_trans3 <= '0';
    end if;
  end if;
  if (BankActivateFlag'EVENT and BankActivateFlag = TRUE) then
    BA := (BA1, BA0);
    RA := (A11, A10, A9, A8, A7, A6, A5, A4, A3, A2, A1, A0); --?
    i := 0;
    j := 0;
    u := 0;
    if (BankActivatedFlag (conv_integer (BA)) = '1') then
      assert false report
      "WARNING : (MEMORY_BANK_ACTIVATE) : Activating same bank without precharge. Command Ignored."
      severity warning;
      BankActivateFinFlag <= transport TRUE, FALSE after 2 ns;
    elsif (BankActivatedFlag (conv_integer (BA)) = '0') then
      if (now - refresh_check(conv_integer(BA), conv_integer(RA)) > 64000000 ns) then
        assert false report
        "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." 
        severity warning;  
      end if;    
      case BA is 
        when "00" =>
          b0_last_activate <= transport now after (tRAS (Part_number) - 1 ns);
          RA_Activated_B0 <= RA;
        when "01" =>
          b1_last_activate <= transport now after (tRAS (Part_number) - 1 ns);
          RA_Activated_B1 <= RA;
        when "10" =>
          b2_last_activate <= transport now after (tRAS (Part_number) - 1 ns);
          RA_Activated_B2 <= RA;
        when "11" =>
          b3_last_activate <= transport now after (tRAS (Part_number) - 1 ns);
          RA_Activated_B3 <= RA;
        when others =>
          assert false report
          "WARNING : (MEMORY_BANK_ACTIVATE) : Impossible Bank Address"
          severity warning;
      end case;
      if (conv_integer (BA) = 0) then
        if (MEM_CELL_ARRAY0(conv_integer (RA)) = NULL) then
          MEM_CELL_ARRAY0(conv_integer (RA)) := NEW ROW_DATA_TYPE;
          loop
            exit when u >= NUM_OF_COLS;
            MEM_CELL_ARRAY0(conv_integer (RA))(u) := 0;
            u := u + 1;
          end loop;
        end if;
        loop
          exit when i >= NUM_OF_COLS;
          TMP_BUF := conv_std_logic_vector(MEM_CELL_ARRAY0 (conv_integer (RA))(i), WORD_SIZE);
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            SA_ARRAY_A0(i, j) <= TMP_BUF(j);
            j := j + 1;
          end loop;
          i := i + 1;
        end loop;
        tmp_act_trans0 <= '1';
      elsif (conv_integer (BA) = 1) then
        if (MEM_CELL_ARRAY1(conv_integer (RA)) = NULL) then
          MEM_CELL_ARRAY1(conv_integer (RA)) := NEW ROW_DATA_TYPE;
          loop
            exit when u >= NUM_OF_COLS;
            MEM_CELL_ARRAY1(conv_integer (RA))(u) := 0;
            u := u + 1;
          end loop;
        end if;
        loop
          exit when i >= NUM_OF_COLS;
          TMP_BUF := conv_std_logic_vector(MEM_CELL_ARRAY1 (conv_integer (RA))(i), WORD_SIZE);
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            SA_ARRAY_A1(i, j) <= TMP_BUF(j);
            j := j + 1;
          end loop;
          i := i + 1;
        end loop;
        tmp_act_trans1 <= '1';
      elsif (conv_integer (BA) = 2) then
        if (MEM_CELL_ARRAY2(conv_integer (RA)) = NULL) then
          MEM_CELL_ARRAY2(conv_integer (RA)) := NEW ROW_DATA_TYPE;
          loop
            exit when u >= NUM_OF_COLS;
            MEM_CELL_ARRAY2(conv_integer (RA))(u) := 0;
            u := u + 1;
          end loop;
        end if;
        loop
          exit when i >= NUM_OF_COLS;
          TMP_BUF := conv_std_logic_vector(MEM_CELL_ARRAY2 (conv_integer (RA))(i), WORD_SIZE);
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            SA_ARRAY_A2(i, j) <= TMP_BUF(j);
            j := j + 1;
          end loop;
          i := i + 1;
        end loop;
        tmp_act_trans2 <= '1';
      elsif (conv_integer (BA) = 3) then
        if (MEM_CELL_ARRAY3(conv_integer (RA)) = NULL) then
          MEM_CELL_ARRAY3(conv_integer (RA)) := NEW ROW_DATA_TYPE;
          loop
            exit when u >= NUM_OF_COLS;
            MEM_CELL_ARRAY3(conv_integer (RA))(u) := 0;
            u := u + 1;
          end loop;
        end if;
        loop
          exit when i >= NUM_OF_COLS;
          TMP_BUF := conv_std_logic_vector(MEM_CELL_ARRAY3 (conv_integer (RA))(i), WORD_SIZE);
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            SA_ARRAY_A3(i, j) <= TMP_BUF(j);
            j := j + 1;
          end loop;
          i := i + 1;
        end loop;
        tmp_act_trans3 <= '1';
      end if;
      BankActivateFinFlag <= transport TRUE, FALSE after 2 ns;
    else
    end if;
  end if;
  if ((PrechargeFlag'EVENT and PrechargeFlag = TRUE) or (AutoPrechargeFlag'EVENT and AutoPrechargeFlag = TRUE)) then
    i := 0;
    j := 0;
    if (PrechargeFlag = TRUE) then
      BA := (BA1, BA0);
    elsif (AutoPrechargeFlag = TRUE) then
      BA := APBA;
    end if;
    if (BA  = "00") then
      b0_last_precharge <= now after 1 ns;
      if (BankActivatedFlag (0) /= '1') then
        assert false report
        "WARNING : (MEMORY_PRECHARGE) : Precharging deactivated bank."
        severity warning;
      else
        RA := RA_Activated_B0;
        SA := SA_ARRAY (conv_integer (BA));
        loop
          exit when i >= NUM_OF_COLS;
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            TMP_BUF(j) := SA(i, j);
            j := j + 1;
          end loop;
          MEM_CELL_ARRAY0 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
          i := i + 1;
        end loop;
        BankActivatedFlag (0) <= '0';
        tmp_ref_addr2 <= RA&"00";
      end if;
    elsif (BA  = "01") then
      b1_last_precharge <= now after 1 ns;
      if (BankActivatedFlag (1) /= '1') then
        assert false report
        "WARNING : (MEMORY_PRECHARGE) : Precharging deactivated bank."
        severity warning;
      else
        RA := RA_Activated_B1;
        SA := SA_ARRAY (conv_integer (BA));
        loop
          exit when i >= NUM_OF_COLS;
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            TMP_BUF(j) := SA(i, j);
            j := j + 1;
          end loop;
          MEM_CELL_ARRAY1 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
          i := i + 1;
        end loop;
        BankActivatedFlag (1) <= '0';
        tmp_ref_addr2 <= RA&"01";
      end if;
    elsif (BA  = "10") then
      b2_last_precharge <= now after 1 ns;
      if (BankActivatedFlag (2) /= '1') then
        assert false report
        "WARNING : (MEMORY_PRECHARGE) : Precharging deactivated bank."
        severity warning;
      else
        RA := RA_Activated_B2;
        SA := SA_ARRAY (conv_integer (BA));
        loop
          exit when i >= NUM_OF_COLS;
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            TMP_BUF(j) := SA(i, j);
            j := j + 1;
          end loop;
          MEM_CELL_ARRAY2 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
          i := i + 1;
        end loop;
        BankActivatedFlag (2) <= '0';
        tmp_ref_addr2 <= RA&"10";
      end if;
    elsif (BA  = "11") then
      b3_last_precharge <= now after 1 ns;
      if (BankActivatedFlag (3) /= '1') then
        assert false report
        "WARNING : (MEMORY_PRECHARGE) : Precharging deactivated bank."
        severity warning;
      else
        RA := RA_Activated_B3;
        SA := SA_ARRAY (conv_integer (BA));
        loop
          exit when i >= NUM_OF_COLS;
          j := 0;
          loop
            exit when j >= WORD_SIZE;
            TMP_BUF(j) := SA(i, j);
            j := j + 1;
          end loop;
          MEM_CELL_ARRAY3 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
          i := i + 1;
        end loop;
        BankActivatedFlag (3) <= '0';
        tmp_ref_addr2 <= RA&"11";
      end if;
    end if;
    if (PrechargeFlag = TRUE) then
      PrechargeFinFlag <= TRUE after 2 ns;
    end if;
  else
    PrechargeFinFlag <= FALSE;
  end if;
  if (PrechargeAllFlag'EVENT and PrechargeAllFlag = TRUE) then
    i := 0;
    if BankActivatedFlag (0) = '1' then
      BA := "00";
      RA := RA_Activated_B0;
      SA := SA_ARRAY (conv_integer (BA));
      loop
        exit when i >= NUM_OF_COLS;
        j := 0;
        loop
          exit when j >= WORD_SIZE;
          TMP_BUF(j) := SA(i, j);
          j := j + 1;
        end loop;
        MEM_CELL_ARRAY0 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
        i := i + 1;
      end loop;
      BankActivatedFlag (0) <= '0';
--    assert false report
--    "NOTE : (PRECHARGE_ALL) : Bank0 Deactivated."
--    severity NOTE;
      tmp_ref_addr2 <= RA&"00";
      b0_last_precharge <= now after 1 ns;
    else
      b0_last_precharge <= now after 1 ns;
    end if;
    if BankActivatedFlag (1) = '1' then
      BA := "01";
      RA := RA_Activated_B1;
      SA := SA_ARRAY (conv_integer (BA));
      loop
        exit when i >= NUM_OF_COLS;
        j := 0;
        loop
          exit when j >= WORD_SIZE;
          TMP_BUF(j) := SA(i, j);
          j := j + 1;
        end loop;
        MEM_CELL_ARRAY1 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
        i := i + 1;
      end loop;
      BankActivatedFlag (1) <= '0';
--    assert false report
--    "NOTE : (PRECHARGE_ALL) : Bank1 Deactivated."
--    severity NOTE;
      tmp_ref_addr2 <= RA&"01";
      b1_last_precharge <= now after 1 ns;
    else
      b1_last_precharge <= now after 1 ns;
    end if;
    if BankActivatedFlag (2) = '1' then
      BA := "10";
      RA := RA_Activated_B2;
      SA := SA_ARRAY (conv_integer (BA));
      loop
        exit when i >= NUM_OF_COLS;
        j := 0;
        loop
          exit when j >= WORD_SIZE;
          TMP_BUF(j) := SA(i, j);
          j := j + 1;
        end loop;
        MEM_CELL_ARRAY2 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
        i := i + 1;
      end loop;
      BankActivatedFlag (2) <= '0';
--    assert false report
--    "NOTE : (PRECHARGE_ALL) : Bank2 Deactivated."
--    severity NOTE;
      tmp_ref_addr2 <= RA&"10";
      b2_last_precharge <= now after 1 ns;
    else
      b2_last_precharge <= now after 1 ns;
    end if;
    if BankActivatedFlag (3) = '1' then
      BA := "11";
      RA := RA_Activated_B3;
      SA := SA_ARRAY (conv_integer (BA));
      loop
        exit when i >= NUM_OF_COLS;
        j := 0;
        loop
          exit when j >= WORD_SIZE;
          TMP_BUF(j) := SA(i, j);
          j := j + 1;
        end loop;
        MEM_CELL_ARRAY3 (conv_integer (RA))(i) := conv_integer (TMP_BUF);
        i := i + 1;
      end loop;
      BankActivatedFlag (3) <= '0';
--    assert false report
--    "NOTE : (PRECHARGE_ALL) : Bank3 Deactivated."
--    severity NOTE;
      tmp_ref_addr2 <= RA&"11";
      b3_last_precharge <= now after 1 ns;
    else
      b3_last_precharge <= now after 1 ns;
    end if;
    if (BankActivatedFlag = "0000") then
      if (PUSCheckFinFlag = TRUE) then
        assert false report
        "WARNING : (PRECHARGE_ALL) : No Activated Banks."
        severity WARNING;
      else
        BankActivatedFlag (0) <= '0';
        BankActivatedFlag (1) <= '0';
        BankActivatedFlag (2) <= '0';
        BankActivatedFlag (3) <= '0';
      end if;
      b0_last_precharge <= now after 1 ns;
      b1_last_precharge <= now after 1 ns;
      b2_last_precharge <= now after 1 ns;
      b3_last_precharge <= now after 1 ns;
    end if;
    PrechargeAllFinFlag <= TRUE;
  else
    PrechargeAllFinFlag <= FALSE;
  end if;
end process; 
----------------------------------------------------------------------------------
--MEMORY_BANK_ACTIVATE_PRECHARGE

----------------------------------------------------------------------------------
------------SENSE_AMPLIFIER_UPDATE
SENSE_AMPLIFIER_UPDATE : process (tmp_act_trans0, tmp_act_trans1, tmp_act_trans2, tmp_act_trans3,
                                  tmp_w_trans0, tmp_w_trans1, tmp_w_trans2, tmp_w_trans3)
begin
  if (tmp_act_trans0'EVENT and tmp_act_trans0 = '1') then 
    SA_ARRAY(0) <= SA_ARRAY_A0; 
  elsif (tmp_act_trans1'EVENT and tmp_act_trans1 = '1') then 
    SA_ARRAY(1) <= SA_ARRAY_A1; 
  elsif (tmp_act_trans2'EVENT and tmp_act_trans2 = '1') then
    SA_ARRAY(2) <= SA_ARRAY_A2; 
  elsif (tmp_act_trans3'EVENT and tmp_act_trans3 = '1') then
    SA_ARRAY(3) <= SA_ARRAY_A3; 
  elsif (tmp_w_trans0'EVENT and tmp_w_trans0 = '1') then 
    SA_ARRAY(0) <= SA_ARRAY_W0; 
  elsif (tmp_w_trans1'EVENT and tmp_w_trans1 = '1') then 
    SA_ARRAY(1) <= SA_ARRAY_W1; 
  elsif (tmp_w_trans2'EVENT and tmp_w_trans2 = '1') then
    SA_ARRAY(2) <= SA_ARRAY_W2; 
  elsif (tmp_w_trans3'EVENT and tmp_w_trans3 = '1') then
    SA_ARRAY(3) <= SA_ARRAY_W3; 
  End if;
end process; 
---------------------------------------------------------------------------------SENSE_AMPLIFIER_UPDATE

DQM_FUNCTION : process (CLK_INT)-------------------------------------------------------------------------
---DQM_FUNCTION

begin
  if (CLK_INT'EVENT and CLK_INT = '1' and (ReadFlag = TRUE or WriteFlag = TRUE)) then
    if (DQM = '1') then
      d_mask1 <= '1';
    end if;
  end if;
  if (CLK_INT'EVENT and CLK_INT = '0') then
    if (d_mask2 = '1') then
      d_mask2 <= transport '0' after 2 ns;
    end if;
    if (d_mask1 = '1') then
      d_mask2 <= transport '1' after 2 ns;
      d_mask1 <= transport '0' after 2 ns;
    end if;
  end if;
end 
process;--------------------------------------------------------------------------
------------------DQM_FUNCTION
    
----------------------------------------------------------------------------------
-----------------------MEMORY_READ
MEMORY_READ : process (PrechargeFlag, PrechargeAllFlag, Read_CA, CLK_INT, CLK_INT_DLY, BurstStopFlag, PCG_ReadFin2)

variable PCG_BA : std_logic_vector ((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable BA : std_logic_vector ((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable CA : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0) := (others => 'X');
variable real_CA : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0) := (others => 'X');
variable AP : std_logic := 'X';
variable DATA : DATA_TYPE;
variable SenseAmpArray : SA_ARRAY_TYPE;
variable CRBI : integer := 0;
variable tmp_CRBI : integer := 0;
variable j : integer := 0;
variable k : integer := 0;
variable CurrentCommand : COMMAND_TYPE := NOP;
variable ChipSelectBar : std_logic := '0';
variable RowAddrStrobeBar : std_logic := '0';
variable ColAddrStrobeBar : std_logic := '0';
variable WriteEnableBar : std_logic := '0';
variable Address10 : std_logic := '0';
variable ClockEnable : CKE_TYPE := (others => '0');
variable BankActFlag : std_logic_vector((NUM_OF_BANKS - 1) downto 0) := (others => '0');

begin
  if ((PrechargeFlag'EVENT and PrechargeFlag = TRUE) or (PrechargeAllFlag'EVENT 
and PrechargeAllFlag = TRUE)) then
    PCG_BA := (BA1, BA0);
    if (((BA = PCG_BA) and (ReadFlag = TRUE) and (PrechargeFlag = TRUE)) or (PrechargeAllFlag = TRUE)) then
      if (PrevModeRegister.CAS_LATENCY = 1) then
        ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentReadBurstIndex <= 0;
        DataQueue(2) <= "ZZZZZZZZ"; --?
        DataQueue(1) <= "ZZZZZZZZ";
        DataQueue(0) <= "ZZZZZZZZ";
      else
        PCG_ReadFin1 <= '1';
      end if;
    end if;
  end if;
  if (Read_CA'EVENT and Read_CA = '1') then
    AP := A10; --?
    BA := (BA1, BA0);
    CA := (A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);  --?
    tmp_CRBI := 0;
    CRBI := 0;
    CurrentReadBurstIndex <= 0;
  end if;
  if (BurstStopFlag'EVENT and BurstStopFlag = TRUE and BurstStopFlag'LAST_VALUE = FALSE and ReadFlag = TRUE) then
    assert false report
    "NOTE : (MEMORY_READ_PROCESS) : Reading terminated by burst stop command."
    severity NOTE;
    CurrentReadBurstIndex <= 0;
    ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
    CurrentReadBurstIndex <= 0;
    DataQueue(2) <= "ZZZZZZZZ";
    DataQueue(1) <= "ZZZZZZZZ";
    DataQueue(0) <= "ZZZZZZZZ";
    DQtransaction <= '0';
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1' and CLK_INT'LAST_VALUE = '0') then
    if (PCG_ReadFin1 = '1') then
      if (PrevModeRegister.CAS_LATENCY = 2) then
        ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentReadBurstIndex <= 0;
        DataQueue(2) <= "ZZZZZZZZ";
        DataQueue(1) <= "ZZZZZZZZ";
        DataQueue(0) <= "ZZZZZZZZ";
        PCG_ReadFin1 <= '0';
      elsif (PrevModeRegister.CAS_LATENCY = 3) then
        PCG_ReadFin2 <= '1';
        PCG_ReadFin1 <= '0';
      end if;
    elsif (PCG_ReadFin2 = '1') then
      if (PrevModeRegister.CAS_LATENCY = 3) then
        ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentReadBurstIndex <= 0;
        DataQueue(2) <= "ZZZZZZZZ";
        DataQueue(1) <= "ZZZZZZZZ";
	if (CurrentState /= READ) THEN
	  DataQueue(0) <= "ZZZZZZZZ";
	end if;
        PCG_ReadFin2 <= '0';
      end if;
    end if;
    if (DQtransaction = '1') Then
       DQtransaction <= '0';
    end if;
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1') then 
    ClockEnable(-1) := CKEN(-1);
    ClockEnable(0) := CKE;
    if ClockEnable (-1) = '1' then
      ChipSelectBar := CSB;
      RowAddrStrobeBar := RASB;
      ColAddrStrobeBar := CASB;
      WriteEnableBar := WEB;
      Address10 := A10;
      BankActFlag := BankActivatedFlag;
    end if;
    COMMAND_DECODE (ChipSelectBar, RowAddrStrobeBar, ColAddrStrobeBar,
      WriteEnableBar, Address10, ClockEnable, CurrentCommand, BankActFlag);
    if(CurrentCommand = RD  and PrevModeRegister.CAS_LATENCY = 3 and
     (PCG_ReadFin1 = '1' or PCG_ReadFin2 = '1' or ReadFinFlag = TRUE)) then
      ReadFix <= 0;
    else
      ReadFix <= ReadFix + 1;
    end if;
  end if;
  if (CLK_INT'EVENT and CLK_INT = '0' and ReadFlag = TRUE) Then
    if (tmp_CRBI >= PrevModeRegister.BURST_LENGTH + PrevModeRegister.CAS_LATENCY - 1) then
      ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
      CurrentReadBurstIndex <= 0;
      DataQueue(2) <= "ZZZZZZZZ";
      DataQueue(1) <= "ZZZZZZZZ";
      DataQueue(0) <= "ZZZZZZZZ";
      CRBI := 0;
      tmp_CRBI := 0;
    end if; 
  end if;
  if (CLK_INT_DLY'EVENT and CLK_INT_DLY = '1' and ReadFlag = TRUE) Then
    DataQueue(2) <= DataQueue(1);
    DataQueue(1) <= DataQueue(0);
    if (CRBI >= PrevModeRegister.BURST_LENGTH) then
      tmp_CRBI := tmp_CRBI + 1;
    end if;
    if (BankActivatedFlag (conv_integer (BA)) = '1') then
      if (CRBI <= PrevModeRegister.BURST_LENGTH - 1) then
        if (WriteFlag = FALSE) then
          SenseAmpArray (conv_integer (BA)):= SA_ARRAY(conv_integer (BA));
          CRBI := CurrentReadBurstIndex;
          tmp_CRBI := CRBI;
          real_CA := real_col_addr;
          MEMORY_READ_ith_ROW (BA, real_CA, DATA, SenseAmpArray);
	  CLKY <= '1';
	  DATA1 <= DATA;
          DataQueue (0) <= DATA;
          CRBI := CRBI + 1;
          tmp_CRBI := CRBI;
          CurrentReadBurstIndex <= CRBI;
	  CLKY <= '0' after 1 ns;
        else
          assert false report
          "NOTE : (MEM_READ_PROCESS) : Read Command is Interrupted by Write Command!"
          severity NOTE;
          DataQueue (0) <= "ZZZZZZZZ";
--EDIT : to fix Read interrupted by Write with DQ masking
          DataQueue (1) <= "ZZZZZZZZ";
          DataQueue (2) <= "ZZZZZZZZ";
          ReadFinFlag <= transport TRUE, FALSE after clk_cycle;
          CRBI := CRBI + 1;
          tmp_CRBI := CRBI;
          CurrentReadBurstIndex <= CRBI;
        end if;
      end if;
    elsif ((CRBI <= PrevModeRegister.BURST_LENGTH - 1) and (PCG_ReadFin1 /= '1' and PCG_ReadFin2 /= '1')) then
      if (WriteFlag = FALSE) then
        assert false report
        "WARNING : (MEM_READ_PROCESS) : Accessing deactivated bank. Garbage DATA may be outputed!"
        severity WARNING;
        DataQueue(0) <= "ZZZZZZZZ";
      else
        assert false report
        "WARNING : (MEM_READ_PROCESS) : And Read Command is Interrupted by Write Command!"
        severity WARNING;
        DataQueue (0) <= "ZZZZZZZZ";
      end if;
      CRBI := CRBI + 1;
      tmp_CRBI := tmp_CRBI + 1;
    else
    end if;
    if (d_mask2 = '1') then
      if (PrevModeRegister.CAS_LATENCY = 1) then
        loop
          exit when j >= WORD_SIZE;
          DataQueue(0)(j) <= 'Z';
          j := j + 1;
        end loop;
      end if;
      j := 0;
      k := 0;
      if (PrevModeRegister.CAS_LATENCY = 2) then
        loop
          exit when j >= WORD_SIZE;
          DataQueue(1)(j) <= 'Z';
          j := j + 1;
        end loop;
      end if;
      j := 0;
      k := 0;
      if (PrevModeRegister.CAS_LATENCY = 3) then
        loop
          exit when j >= WORD_SIZE;
          DataQueue(2)(j) <= 'Z';
          j := j + 1;
        end loop;
      end if; 
      j := 0;
      k := 0;
    end if;
    DQtransaction <= '1';    
  else
  end if;
end process; 
----------------------------------------------------------------------------------
----------MEMORY_READ

----------------------------------------------------------------------------------
------------DATA_PIPE_LINE_CONTROL
DATA_PIPE_LINE_CONTROL : process (ReadFlag, Read_CA, DQtransaction, BurstStopFlag)

variable i : integer := 0;

begin
  if (CLKStableFlag /= TRUE) then
    DQ0 <= 'Z';
    DQ1 <= 'Z';
    DQ2 <= 'Z';
    DQ3 <= 'Z';
    DQ4 <= 'Z';
    DQ5 <= 'Z';
    DQ6 <= 'Z';
    DQ7 <= 'Z';
  end if;
  if (BurstStopFlag = TRUE) then
    DQ0 <= transport 'Z' after 2 ns;
    DQ1 <= transport 'Z' after 2 ns;
    DQ2 <= transport 'Z' after 2 ns;
    DQ3 <= transport 'Z' after 2 ns;
    DQ4 <= transport 'Z' after 2 ns;
    DQ5 <= transport 'Z' after 2 ns;
    DQ6 <= transport 'Z' after 2 ns;
    DQ7 <= transport 'Z' after 2 ns;
  elsif (DQtransaction'EVENT and DQtransaction = '1') then
    i := i + 1; 
    case PrevModeRegister.CAS_LATENCY is
      when 1 =>
        If (i = 1) Then
            DQ0 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(0) after (tAC1 (Part_Number) - 2 ns);
            DQ1 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(1) after (tAC1 (Part_Number) - 2 ns);
            DQ2 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(2) after (tAC1 (Part_Number) - 2 ns);
            DQ3 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(3) after (tAC1 (Part_Number) - 2 ns);
            DQ4 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(4) after (tAC1 (Part_Number) - 2 ns);
            DQ5 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(5) after (tAC1 (Part_Number) - 2 ns);
            DQ6 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(6) after (tAC1 (Part_Number) - 2 ns);
            DQ7 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(0)(7) after (tAC1 (Part_Number) - 2 ns);
          i := i - 1;
        End if;
      when 2 =>
        if (i = 2) Then
            DQ0 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(0) after (tAC2 (Part_Number) - 2 ns);
            DQ1 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(1) after (tAC2 (Part_Number) - 2 ns);
            DQ2 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(2) after (tAC2 (Part_Number) - 2 ns);
            DQ3 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(3) after (tAC2 (Part_Number) - 2 ns);
            DQ4 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(4) after (tAC2 (Part_Number) - 2 ns);
            DQ5 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(5) after (tAC2 (Part_Number) - 2 ns);
            DQ6 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(6) after (tAC2 (Part_Number) - 2 ns);
            DQ7 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(1)(7) after (tAC2 (Part_Number) - 2 ns);
          i := i - 1;
        End if;
      when 3 =>
        if (i = 3) Then -- try to fix a non read problem
--        if (i = 3 and ReadFix /= 1) Then
            DQ0 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(0) after (tAC3 (Part_Number) - 2 ns);
            DQ1 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(1) after (tAC3 (Part_Number) - 2 ns);
            DQ2 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(2) after (tAC3 (Part_Number) - 2 ns);
            DQ3 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(3) after (tAC3 (Part_Number) - 2 ns);
            DQ4 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(4) after (tAC3 (Part_Number) - 2 ns);
            DQ5 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(5) after (tAC3 (Part_Number) - 2 ns);
            DQ6 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(6) after (tAC3 (Part_Number) - 2 ns);
            DQ7 <= transport 'Z' after (tOH (Part_Number) - 2 ns),
                             DataQueue(2)(7) after (tAC3 (Part_Number) - 2 ns);
          i := i - 1;
	else
          DQ0 <= transport 'Z' after tOH (Part_Number);
    	  DQ1 <= transport 'Z' after tOH (Part_Number);
  	  DQ2 <= transport 'Z' after tOH (Part_Number);
          DQ3 <= transport 'Z' after tOH (Part_Number);
          DQ4 <= transport 'Z' after tOH (Part_Number);
          DQ5 <= transport 'Z' after tOH (Part_Number);
          DQ6 <= transport 'Z' after tOH (Part_Number);
          DQ7 <= transport 'Z' after tOH (Part_Number);
	  if(i = 3) then
	    i := i - 1;
	  end if;
        End if;
      when others =>
    end case;
  elsif (DQtransaction'EVENT and DQtransaction = '0') then
    DQ0 <= transport 'Z' after tOH (Part_Number);
    DQ1 <= transport 'Z' after tOH (Part_Number);
    DQ2 <= transport 'Z' after tOH (Part_Number);
    DQ3 <= transport 'Z' after tOH (Part_Number);
    DQ4 <= transport 'Z' after tOH (Part_Number);
    DQ5 <= transport 'Z' after tOH (Part_Number);
    DQ6 <= transport 'Z' after tOH (Part_Number);
    DQ7 <= transport 'Z' after tOH (Part_Number);
  end if;
  if (ReadFlag'EVENT and ReadFlag = FALSE) then
    i := 0;
  end if;
end 
process;--------------------------------------------------------------------------
--------DATA_PIPE_LINE_CONTROL

BURST_RD_WR_ADDR_GEN : process(CLK_INT, Read_CA, Write_CA, ReadFlag, WriteFlag)--------------------BURST_RD_WR_ADDR_GEN

variable CA : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0) := (others => 'X'); --?
variable i, j : integer := 0;

begin
  if ((ReadFlag'EVENT and ReadFlag = TRUE) or (WriteFlag'EVENT and WriteFlag = TRUE)) then
    real_col_addr <= "0000000000";
    i := 0;
    j := 0;
  end if;
  if ((Read_CA'EVENT and Read_CA = '1') or (Write_CA'EVENT and Write_CA = '1')) then
    PrevModeRegister <= ModeRegister;
    CA := (A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);
    i := 0;
    j := 0;
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1' and CLK_INT'LAST_VALUE = '0') then
    if (WriteFlag = TRUE or ReadFlag = TRUE) then
      if (Read_CA = '1' or Write_CA = '1') then
        real_col_addr <= (A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);
      elsif (PrevModeRegister.BURST_LENGTH = 1) then
        real_col_addr <= CA;
      elsif (PrevModeRegister.BURST_LENGTH = 2) then
        real_col_addr <= CA((NUM_OF_COL_ADD - 1) downto 1)& 
                         conv_std_logic_vector(remainder((conv_integer (CA(0)) + i), 2), 1);
      elsif (PrevModeRegister.BURST_LENGTH = 4) then
        if (PrevModeRegister.BURST_MODE = SEQUENTIAL) then
          real_col_addr <= CA((NUM_OF_COL_ADD - 1) downto 2)& 
                           conv_std_logic_vector(remainder((conv_integer(CA(1 downto 0)) + i), 4), 2);
        elsif (PrevModeRegister.BURST_MODE = INTERLEAVE) then
          real_col_addr <= CA((NUM_OF_COL_ADD - 1) downto 2)& 
                           xor_func(CA(1 downto 0), conv_std_logic_vector(i, 2)); 
        end if;
      elsif (PrevModeRegister.BURST_LENGTH = 8) then
        if (PrevModeRegister.BURST_MODE = SEQUENTIAL) then
          real_col_addr <= CA((NUM_OF_COL_ADD - 1) downto 3)& 
                           conv_std_logic_vector(remainder((conv_integer(CA(2 downto 0)) + i), 8), 3);
        elsif (PrevModeRegister.BURST_MODE = INTERLEAVE) then
          real_col_addr <= CA((NUM_OF_COL_ADD - 1) downto 3)& 
                           xor_func(CA(2 downto 0), conv_std_logic_vector(i, 3));
        end if;
      elsif (PrevModeRegister.BURST_LENGTH = NUM_OF_COLS) then 
        if (PrevModeRegister.BURST_MODE = SEQUENTIAL) then
          real_col_addr <= conv_std_logic_vector(remainder(conv_integer(CA((NUM_OF_COL_ADD - 1) downto 0) + i), NUM_OF_COLS), NUM_OF_COL_ADD); 
        end if;
      end if;
      i := i + 1;
    else
      real_col_addr <= "0000000000"; --?
      i := 0;
      j := 0;
    end if;
  end if;   
end 
process;--------------------------------------------------------------------------
----------BURST_RD_WR_ADDR_GEN

----------------------------------------------------------------------------------
----------------------MEMORY_WRITE
MEMORY_WRITE : process (Read_CA, PrechargeFlag, PrechargeAllFlag, Write_CA, CLK_INT, BurstStopFlag)

variable PCG_BA : std_logic_vector ((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable BA : std_logic_vector ((NUM_OF_BANK_ADD - 1) downto 0) := (others => 'X');
variable CA : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0) := (others => 'X');
variable real_CA : std_logic_vector ((NUM_OF_COL_ADD - 1) downto 0) := (others => 'X');
variable Din : std_logic_vector ((WORD_SIZE - 1) downto 0) := (others => 'X');
variable WriteDriverArray : SA_ARRAY_TYPE;
variable SenseAmpArray : SA_ARRAY_TYPE;
variable AP : std_logic := '0';
variable A, i, j, k, l, m : integer := 0;
variable CWBI : integer := 0;

begin
  if (PrechargeFlag'EVENT and PrechargeFlag = TRUE and PrechargeFlag'LAST_VALUE = FALSE) then
    PCG_BA := (BA1, BA0);
    if ((BA = PCG_BA) and (WriteFlag = TRUE)) then
      PCG_WriteFin <= transport '1', '0' after 2 ns;
    end if;
  end if;
  if (PrechargeAllFlag'EVENT and PrechargeAllFlag = TRUE and PrechargeAllFlag'LAST_VALUE = FALSE and WriteFlag = TRUE) then
      PCG_WriteFin <= transport '1', '0' after 2 ns;
  end if;  
  if (Write_CA'EVENT and Write_CA = '1') then
    BA := (BA1, BA0);
--    WriteDriverArray (conv_integer (BA)):= SA_ARRAY (conv_integer (BA));
--    SenseAmpArray (conv_integer (BA)):= SA_ARRAY (conv_integer (BA));
    CA := (A9, A8, A7, A6, A5, A4, A3, A2, A1, A0);
    AP := A10;
    CurrentWriteBurstIndex <= 0;
    CWBI := 0;
  end if;
  if (BurstStopFlag'EVENT and BurstStopFlag = TRUE and BurstStopFlag'LAST_VALUE = FALSE and WriteFlag = TRUE) then
    assert false report
    "NOTE : (MEM_WRITE_PROCESS) : Writing terminated by burst stop command."
    severity NOTE;
    bstop_write <= '1';
    CurrentWriteBurstIndex <= 0;
  end if;
  if (Read_CA'EVENT and Read_CA = '1') then
    WriteFinFlag <= transport TRUE, FALSE after clk_cycle;
    CurrentWriteBurstIndex <= 0;
    CWBI := 0;
  end if;
  if ((CLK_INT'EVENT and CLK_INT = '1') or (Write_CA'EVENT and Write_CA = '1')) Then
    Din := (DQ7, DQ6, DQ5, DQ4, DQ3, DQ2, DQ1, DQ0);
  End if;
  if (CLK_INT'EVENT and CLK_INT = '1' and ((WriteFlag = TRUE) or ((WriteFlag = FALSE) and (tmp_w_trans0 = '1')))) Then
    tmp_w_trans0 <= '0';
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1' and ((WriteFlag = TRUE) or ((WriteFlag = FALSE) and (tmp_w_trans1 = '1')))) Then
    tmp_w_trans1 <= '0';
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1' and ((WriteFlag = TRUE) or ((WriteFlag = FALSE) and (tmp_w_trans2 = '1')))) Then
    tmp_w_trans2 <= '0';
  end if;
  if (CLK_INT'EVENT and CLK_INT = '1' and ((WriteFlag = TRUE) or ((WriteFlag = FALSE) and (tmp_w_trans3 = '1')))) Then
    tmp_w_trans3 <= '0';
  end if;
  if (CLK_INT'EVENT and CLK_INT = '0' and WriteFlag = TRUE) then
    if (BankActivatedFlag (conv_integer(BA)) = '1') then
      WriteDriverArray(conv_integer (BA)) := SA_ARRAY (conv_integer (BA));
      if (conv_integer(BA) = 0) then
        SA_ARRAY_W0 <= SA_ARRAY (conv_integer (BA));
      elsif (conv_integer(BA) = 1) then
        SA_ARRAY_W1 <= SA_ARRAY (conv_integer (BA));
      elsif (conv_integer(BA) = 2) then
        SA_ARRAY_W2 <= SA_ARRAY (conv_integer (BA));
      elsif (conv_integer(BA) = 3) then
        SA_ARRAY_W3 <= SA_ARRAY (conv_integer (BA));
      end if;
      CWBI := CurrentWriteBurstIndex;
      if (d_mask1 = '0' and ReadFlag = FALSE) then 
        loop
          real_CA := real_col_addr;
	  real_CA1 <= real_col_addr;
          WriteDriverArray(conv_integer(BA))(conv_integer (real_CA), k) := Din (k);
          k := k + 1;
          exit when (k >= WORD_SIZE/2);
        end loop;
        real_CA := real_col_addr;
        MEMORY_WRITE_a_ROW (BA, real_CA, WriteDriverArray, SenseAmpArray);
        l := 0;
        loop
          real_CA := real_col_addr;
          if (conv_integer (BA) = 0) then
            SA_ARRAY_W0 (conv_integer(real_CA), l) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), l);
          elsif (conv_integer (BA) = 1) then
            SA_ARRAY_W1 (conv_integer(real_CA), l) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), l);
          elsif (conv_integer (BA) = 2) then
            SA_ARRAY_W2 (conv_integer(real_CA), l) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), l);
          elsif (conv_integer (BA) = 3) then
            SA_ARRAY_W3 (conv_integer(real_CA), l) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), l);
          end if;
          l := l + 1;
          exit when (l >= WORD_SIZE/2);
        end loop;
      last_data_in <= now;
      end if;
      if (d_mask1 = '0' and ReadFlag = FALSE) then
        m := (WORD_SIZE-1);
        loop
          real_CA := real_col_addr;
	  real_CA1 <= real_col_addr;
          WriteDriverArray(conv_integer(BA))(conv_integer (real_CA), m) := Din (m);
	  Din1 <= Din;
	  M1 <= m;
          m := m - 1;
          exit when (m < WORD_SIZE/2);
        end loop;
        real_CA := real_col_addr;
        MEMORY_WRITE_a_ROW (BA, real_CA, WriteDriverArray, SenseAmpArray);
        i := (WORD_SIZE-1);
        loop
          real_CA := real_col_addr;
          if (conv_integer (BA) = 0) then
            SA_ARRAY_W0 (conv_integer(real_CA), i) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), i);
          elsif (conv_integer (BA) = 1) then
            SA_ARRAY_W1 (conv_integer(real_CA), i) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), i);
          elsif (conv_integer (BA) = 2) then
            SA_ARRAY_W2 (conv_integer(real_CA), i) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), i);
          elsif (conv_integer (BA) = 3) then
            SA_ARRAY_W3 (conv_integer(real_CA), i) <= SenseAmpArray (conv_integer(BA))(conv_integer(real_CA), i);
          end if;
          i := i - 1;
          exit when (i < WORD_SIZE/2);
        end loop;
      last_data_in <= now;
      end if; 
      if (conv_integer (BA) = 0) then
        tmp_w_trans0 <= '1';
      elsif (conv_integer (BA) = 1) then
        tmp_w_trans1 <= '1';
      elsif (conv_integer (BA) = 2) then
        tmp_w_trans2 <= '1';
      elsif (conv_integer (BA) = 3) then
        tmp_w_trans3 <= '1';
      end if;
      CWBI := CWBI + 1;
      if (PrevModeRegister.WRITE_MODE = BRSW) then
        WriteFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentWriteBurstIndex <= 0;
        CWBI := 0;
      elsif (CWBI >= PrevModeRegister.BURST_LENGTH) then
        WriteFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentWriteBurstIndex <= 0;
        CWBI := 0;
      elsif (CWBI < PrevModeRegister.BURST_LENGTH) then
        if (BA = "00") then
          b0_last_data_in <= (now - clk_cycle/2);
        elsif (BA = "01") then
          b1_last_data_in <= (now - clk_cycle/2);
        elsif (BA = "10") then
          b2_last_data_in <= (now - clk_cycle/2);
        elsif (BA = "11") then
          b3_last_data_in <= (now - clk_cycle/2);
        end if;
      end if;
      CurrentWriteBurstIndex <= CWBI;
      j := 0;
      k := 0;
      l := 0;
      m := 0;
    else
      assert false report
      "WARNING : (MEM_WRITE_PROCESS) : Accessing deactivated bank."
      severity WARNING;
      CWBI := CWBI + 1;
      if (PrevModeRegister.WRITE_MODE = BRSW) then
        WriteFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentWriteBurstIndex <= 0;
        CWBI := 0;
      elsif (CWBI >= PrevModeRegister.BURST_LENGTH) then
        WriteFinFlag <= transport TRUE, FALSE after clk_cycle;
        CurrentWriteBurstIndex <= 0;
        CWBI := 0;
      end if;
    end if;
    A := (WORD_SIZE-1);
    loop
      WRITEDRIVEARRAR_CHECK0(A) <= WriteDriverArray(conv_integer(BA))(conv_integer (real_CA),A);
      WRITEDRIVEARRAR_CHECK1(A) <= WriteDriverArray(conv_integer(BA))(conv_integer (real_CA),A);
      WRITEDRIVEARRAR_CHECK2(A) <= WriteDriverArray(conv_integer(BA))(conv_integer (real_CA),A);
      A := A - 1;
      exit when (A < 0);
    end loop;
  else
  end if;
end process; 
----------------------------------------------------------------------------------
---------MEMORY_WRITE

----------------------------------------------------------------------------------
-----------REFRESH_COUNTER_PROCESS
REFRESH_COUNTER : PROCESS(AutoRefFlag, SelfRefFlag, CLK) 

variable rc : integer := 0;
variable i, j : integer := 0;
variable tmp_time : time := 0 ns;

begin
  if (AutoRefFlag'EVENT and AutoRefFlag = TRUE) then
    if (rc >= 4096) then --? 4K or 8K
      rc := rc - 4096;  --?
    end if;
    RefCnt <= conv_std_logic_vector (rc, NUM_OF_ROW_ADD);
    RefCnt_trans <= '1', '0' after 3 ns;
    if (now - refresh_check(0, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 0))) > 64000000 ns) then
      assert false report
      "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." 
      severity warning;
    end if;
    if (now - refresh_check(1, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 0))) > 64000000 ns) then
      assert false report
      "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." 
      severity warning;
    end if;
    if (now - refresh_check(2, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 0))) > 64000000 ns) then
      assert false report
      "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." 
      severity warning;
    end if;
    if (now - refresh_check(3, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 0))) > 64000000 ns) then
      assert false report
      "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." 
      severity warning;
    end if;
    tmp_ref_addr1 <= conv_std_logic_vector (rc, NUM_OF_ROW_ADD);
    rc := rc + 1;
    last_aref_time <= now;
  end if;
  if (CLK'EVENT and CLK = '1') then
    if (SelfRefFlag = TRUE and SelfRefExtFlag /= TRUE) then
      Ref_time <= Ref_time + clk_cycle; 
      if (Ref_time >= (15600 ns - clk_cycle)) then
        Ref_time <= 0 ns;
        if (rc >= 4096) then
          rc := rc - 4096;
        end if;
        RefCnt <= conv_std_logic_vector (rc, NUM_OF_ROW_ADD);
        RefCnt_trans <= '1', '0' after 3 ns;
        if (now - refresh_check(0, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 1))) > 64000000 ns) then
          assert false report
          "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." --?
          severity warning;
        end if;
        if (now - refresh_check(1, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 1))) > 64000000 ns) then
          assert false report
          "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." --?
          severity warning;
        end if;
        if (now - refresh_check(2, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 1))) > 64000000 ns) then
          assert false report
          "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." --?
          severity warning;
        end if;
        if (now - refresh_check(3, conv_integer(RefCnt((NUM_OF_ROW_ADD-1) downto 1))) > 64000000 ns) then
          assert false report
          "WARNING : (REFRESH_INTERVAL) : Refresh Interval Exceeds 64ms. So, This Row's Data Is Lost." --?
          severity warning;
        end if;
        tmp_ref_addr1 <= conv_std_logic_vector (rc, NUM_OF_ROW_ADD);
        rc := rc + 1;
      end if;
    end if;
  end if;
end 
process;--------------------------------------------------------------------------
-------REFRESH_COUNTER_PROCESS

----------------------------------------------------------------------------------
-----------POWER_UP_SEQUENCE_CHECK
PUS_CHECK : process(CLK, PrechargeAllFlag, AutoRefFlag)

variable ChipSelectBar : std_logic := '0';
variable RowAddrStrobeBar : std_logic := '0';
variable ColAddrStrobeBar : std_logic := '0';
variable WriteEnableBar : std_logic := '0';
variable Address10 : std_logic := '0';
variable ClockEnable : CKE_TYPE := (others => '0');
variable CurrentCommand : COMMAND_TYPE := NOP;
variable i, j : integer := 0;
variable PUSNOP200USFlag : boolean := FALSE;
variable PUSPCGAFlag : boolean := FALSE;
variable PUSAREF8Flag : boolean := FALSE;
variable BankActFlag : std_logic_vector((NUM_OF_BANKS - 1) downto 0) := (others => '0');
variable pus_aref : integer := 0;

begin
  if ((Clk_Cycle_Chk = TRUE) and (PUSCheckFinFlag /= TRUE)) then
    if (CLK'EVENT and CLK = '1') then
      cur_time <= now;
    end if;
    if (PrechargeAllFlag'EVENT and PrechargeAllFlag = TRUE) then
      assert (cur_time >= tPUS - clk_cycle) report
      "ERROR : (Power Up Sequence Check) : Input Stable (for 200us) error !"
      severity error; 
    end if;
    if (AutoRefFlag'EVENT and AutoRefFlag = TRUE) then
      assert (PUSPCGAFlag = TRUE) report
      "ERROR : (Power Up Sequence Check) : Needs Precharge All Bank Command before Auto Refresh !"
      severity error;
      pus_aref := pus_aref + 1;
      if (pus_aref >= 8) then
        PUSAREF8Flag := TRUE;
      end if;
    end if;  
    ClockEnable := CKEN;
    if ClockEnable (-1) = '1' then
      ChipSelectBar := CSB;
      RowAddrStrobeBar := RASB;
      ColAddrStrobeBar := CASB;
      WriteEnableBar := WEB;
      Address10 := A10;
      BankActFlag := BankActivatedFlag;
    end if;
    COMMAND_DECODE (ChipSelectBar, RowAddrStrobeBar, ColAddrStrobeBar,
                    WriteEnableBar, Address10, ClockEnable, CurrentCommand, BankActFlag);
    if ((PUSNOP200USFlag = FALSE) and (PUSAREF8Flag /= TRUE)) then
      if (CLK = '1') then
        i := i + 1;
        if (i * clk_cycle >= tPUS - clk_cycle) then
          PUSNOP200USFlag := TRUE;
        end if;
      end if;
    end if;
    if ((CurrentCommand = PCGA) and (PUSNOP200USFlag = TRUE) and (PUSAREF8Flag /= TRUE)) then
      PUSPCGAFlag := TRUE;
    elsif ((CurrentCommand = AREF) and (PUSNOP200USFlag = TRUE) and (PUSAREF8Flag = TRUE)) then
      PUSCheckFinFlag <= TRUE;
    end if;
  end if;
end 
process;--------------------------------------------------------------------------
-------POWER_UP_SEQUENCE_CHECK

----------------------------------------------------------------------------------
-----------------BUS_CHANGE_DETECT
BUS_CHANGE_DETECT : process(CKE, WEB, CASB, RASB, CSB, DQM, A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11,
                            BA0, BA1, DQ0, DQ1, DQ2, DQ3, DQ4, DQ5, DQ6, DQ7)
begin
  if (CLKStableFlag = TRUE) then
    if (CKE'EVENT and CKE /= CKE'LAST_VALUE) then
      cke_ch <= now after 0.1 ns; end if;
    if (WEB'EVENT and WEB /= WEB'LAST_VALUE) then
      web_ch <= now after 0.1 ns; end if;
    if (CASB'EVENT and CASB /= CASB'LAST_VALUE) then
      casb_ch <= now after 0.1 ns; end if;
    if (RASB'EVENT and RASB /= RASB'LAST_VALUE) then
      rasb_ch <= now after 0.1 ns; end if;
    if (CSB'EVENT and CSB /= CSB'LAST_VALUE) then
      csb_ch <= now after 0.1 ns; end if;
    if (DQM'EVENT and DQM /= DQM'LAST_VALUE) then
      dqm_ch <= now after 0.1 ns; end if;
    if (A0'EVENT and A0 /= A0'LAST_VALUE) then
      a0_ch <= now after 0.1 ns; end if;
    if (A1'EVENT and A1 /= A1'LAST_VALUE) then
      a1_ch <= now after 0.1 ns; end if;
    if (A2'EVENT and A2 /= A2'LAST_VALUE) then
      a2_ch <= now after 0.1 ns; end if;
    if (A3'EVENT and A3 /= A3'LAST_VALUE) then
      a3_ch <= now after 0.1 ns; end if;
    if (A4'EVENT and A4 /= A4'LAST_VALUE) then
      a4_ch <= now after 0.1 ns; end if;
    if (A5'EVENT and A5 /= A5'LAST_VALUE) then
      a5_ch <= now after 0.1 ns; end if;
    if (A6'EVENT and A6 /= A6'LAST_VALUE) then
      a6_ch <= now after 0.1 ns; end if;
    if (A7'EVENT and A7 /= A7'LAST_VALUE) then
      a7_ch <= now after 0.1 ns; end if;
    if (A8'EVENT and A8 /= A8'LAST_VALUE) then
      a8_ch <= now after 0.1 ns; end if;
    if (A9'EVENT and A9 /= A9'LAST_VALUE) then
      a9_ch <= now after 0.1 ns; end if;
    if (A10'EVENT and A10 /= A10'LAST_VALUE) then
      a10_ch <= now after 0.1 ns; end if;
    if (A11'EVENT and A11 /= A11'LAST_VALUE) then
      a11_ch <= now after 0.1 ns; end if;
    if (BA0'EVENT and BA0 /= BA0'LAST_VALUE) then
      ba0_ch <= now after 0.1 ns; end if;
    if (BA1'EVENT and BA1 /= BA1'LAST_VALUE) then
      ba1_ch <= now after 0.1 ns; end if;
    if (DQ0'EVENT and STD_LOGIC_TO_BIT (DQ0) /= STD_LOGIC_TO_BIT (DQ0'LAST_VALUE)) then
      dq0_ch <= now after 0.1 ns; end if;
    if (DQ1'EVENT and STD_LOGIC_TO_BIT (DQ1) /= STD_LOGIC_TO_BIT (DQ1'LAST_VALUE)) then
      dq1_ch <= now after 0.1 ns; end if;
    if (DQ2'EVENT and STD_LOGIC_TO_BIT (DQ2) /= STD_LOGIC_TO_BIT (DQ2'LAST_VALUE)) then
      dq2_ch <= now after 0.1 ns; end if;
    if (DQ3'EVENT and STD_LOGIC_TO_BIT (DQ3) /= STD_LOGIC_TO_BIT (DQ3'LAST_VALUE)) then
      dq3_ch <= now after 0.1 ns; end if;
    if (DQ4'EVENT and STD_LOGIC_TO_BIT (DQ4) /= STD_LOGIC_TO_BIT (DQ4'LAST_VALUE)) then
      dq4_ch <= now after 0.1 ns; end if;
    if (DQ5'EVENT and STD_LOGIC_TO_BIT (DQ5) /= STD_LOGIC_TO_BIT (DQ5'LAST_VALUE)) then
      dq5_ch <= now after 0.1 ns; end if;
    if (DQ6'EVENT and STD_LOGIC_TO_BIT (DQ6) /= STD_LOGIC_TO_BIT (DQ6'LAST_VALUE)) then
      dq6_ch <= now after 0.1 ns; end if;
    if (DQ7'EVENT and STD_LOGIC_TO_BIT (DQ7) /= STD_LOGIC_TO_BIT (DQ7'LAST_VALUE)) then
      dq7_ch <= now after 0.1 ns; end if;
  else
  end if;
end 
process;--------------------------------------------------------------------------
-------------BUS_CHANGE_DETECT

----------------------------------------------------------------------------------
------------------------CLK_TIMING
CLK_TIMING : process (CLK_INT)
  variable clk_last_cycle : time := 0 ns;
  variable i : integer := 0;
begin
  if (CLK_INT='1') then
    if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
      tCCD <= clk_cycle;
      tDPL <= clk_cycle;
      tDAL <= tDPL + tRP (Part_Number);
      tMRD <= 2 * clk_cycle;
      tPDE <= clk_cycle;
      tSRE <= clk_cycle;
      if (ModeRegister.CAS_LATENCY = 3) then
        assert (clk_cycle >= tCK3 (Part_Number)) report
        "ERROR : (CLK_TIMING) : Clock period is too small.(tCK3)"
        severity error;
      elsif (ModeRegister.CAS_LATENCY = 2) then
        assert (clk_cycle >= tCK2 (Part_Number)) report
        "ERROR : (CLK_TIMING) : Clock period is too small.(tCK2)"
        severity error;
      elsif (ModeRegister.CAS_LATENCY = 1) then
        assert (clk_cycle >= tCK1) report
        "ERROR : (CLK_TIMING) : Clock period is too small.(tCK1)"
        severity error;
      else
        assert (clk_cycle >= tck2 (Part_Number)) report
        "ERROR : (CLK_TIMING) : Mode register read error."
         severity error;
      end if;
      assert (now - clk_last_falling >= tCHW (Part_Number)) report
      "ERROR : (CLK_TIMING) : Clock high time is too small."
      severity error;
    end if;
    clk_last_rising <= now after 1 ns;
  end if;
  if (CLK_INT = '0') then
    clk_last_falling <= now after 1 ns;
    if (CLKStableFlag = TRUE) then
      assert (now - clk_last_rising >= tCLW (Part_Number)) report
      "ERROR : (CLK_TIMING) : Clock low time is too small."
      severity error;
    end if;
  end if;
  if (clk_last_cycle - clk_cycle <= 0.1 * clk_last_cycle) then
    if (i >= 100) then
      CLKStableFlag <= TRUE;
    else
--    assert false report "NOTE : Waiting until CLK stabilized." severity note;
    end if;
    clk_last_cycle := clk_cycle;
    i := i + 1;
  end if;
end 
process;--------------------------------------------------------------------------
--------------------CLK_TIMING

----------------------------------------------------------------------------------
-----------------------------SETUP
SETUP : process (CLK_INT, WriteFlag)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if ((WriteFlag'EVENT and WriteFlag = TRUE) or (CLK_INT'EVENT and CLK_INT = '1' and WriteFlag = TRUE)) then
    assert (
      ((now - dq0_ch) >= tDS (Part_Number)) and
      ((now - dq1_ch) >= tDS (Part_Number)) and
      ((now - dq2_ch) >= tDS (Part_Number)) and
      ((now - dq3_ch) >= tDS (Part_Number)) and
      ((now - dq4_ch) >= tDS (Part_Number)) and
      ((now - dq5_ch) >= tDS (Part_Number)) and
      ((now - dq6_ch) >= tDS (Part_Number)) and
      ((now - dq7_ch) >= tDS (Part_Number))) report
    "ERROR : (SETUP_HOLD) : Data input setup time violation."
    severity error;
  end if;
  if (CLK_INT'EVENT and (CLK_INT = '1')) then
    assert (
      ((now - a0_ch) >= tAS (Part_Number)) and
      ((now - a1_ch) >= tAS (Part_Number)) and
      ((now - a2_ch) >= tAS (Part_Number)) and
      ((now - a3_ch) >= tAS (Part_Number)) and
      ((now - a4_ch) >= tAS (Part_Number)) and
      ((now - a5_ch) >= tAS (Part_Number)) and
      ((now - a6_ch) >= tAS (Part_Number)) and
      ((now - a7_ch) >= tAS (Part_Number)) and
      ((now - a8_ch) >= tAS (Part_Number)) and
      ((now - a9_ch) >= tAS (Part_Number)) and
      ((now - a10_ch) >= tAS (Part_Number)) and
      ((now - a11_ch) >= tAS (Part_Number)) and
      ((now - ba0_ch) >= tAS (Part_Number)) and
      ((now - ba1_ch) >= tAS (Part_Number))) report
    "ERROR : (SETUP_HOLD) : Address input setup time violation."
    severity error;
    assert (
      ((now - csb_ch) >= tCS (Part_Number)) and
      ((now - rasb_ch) >= tCS (Part_Number)) and
      ((now - casb_ch) >= tCS (Part_Number)) and
      ((now - web_ch) >= tCS (Part_Number)) and
      ((now - dqm_ch) >= tCS (Part_Number))) report
    "ERROR : (SETUP_HOLD) : Command input setup time violation."
    severity error;
    assert (now - cke_ch >= tCKS (Part_Number)) report
    "ERROR : (SETUP_HOLD) : Clock enable input setup time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
------------------------SETUP

----------------------------------------------------------------------------------
---------------------------A_HOLD
A_HOLD : process (A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11, BA0, BA1)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  assert ((now - clk_last_rising) >= tAH (Part_Number)) report
  "ERROR : (SETUP_HOLD) : Address hold time violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
-----------------------A_HOLD

----------------------------------------------------------------------------------
--------------------------DQ_HOLD
DQ_HOLD : process (DQ0, DQ1, DQ2, DQ3, DQ4, DQ5, DQ6, DQ7)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  assert ((now - clk_last_rising) >= tDH (Part_Number)) report
  "ERROR : (SETUP_HOLD) : Data hold time violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
----------------------DQ_HOLD

----------------------------------------------------------------------------------
---------------------------C_HOLD
C_HOLD : process (CSB, WEB, RASB, CASB)
begin
if (CLKStableFlag = TRUE) then
  assert ((now - clk_last_rising) >= tCH (Part_Number)) report
  "ERROR : (SETUP_HOLD) : Command hold time violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
-----------------------C_HOLD

----------------------------------------------------------------------------------
-------------------------CKE_HOLD
CKE_HOLD : process (CKE)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  assert ((now - clk_last_rising) >= tCKH (Part_Number)) report
  "ERROR : (SETUP_HOLD) : Clock enable hold time violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
---------------------CKE_HOLD

----------------------------------------------------------------------------------
------------------------tRC_CHECK
tRC_CHECK : process (BankActivateFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b0_last_activate) >= tRC (Part_Number)) report
    "ERROR : (tRC_CHECK) : Row Address Strobe cycle time violation."
    severity error;
  elsif (BA = "01" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b1_last_activate) >= tRC (Part_Number)) report
    "ERROR : (tRC_CHECK) : Row Address Strobe cycle time violation."
    severity error;
  elsif (BA = "10" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b2_last_activate) >= tRC (Part_Number)) report
    "ERROR : (tRC_CHECK) : Row Address Strobe cycle time violation."
    severity error;
  elsif (BA = "11" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b3_last_activate) >= tRC (Part_Number)) report
    "ERROR : (tRC_CHECK) : Row Address Strobe cycle time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
--------------------tRC_CHECK

----------------------------------------------------------------------------------
-----------------------tRCD_CHECK
tRCD_CHECK : process (ReadFlag, WriteFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and ((ReadFlag = TRUE) or (WriteFlag = TRUE))) then
    assert ((now - b0_last_activate) >= tRCD (Part_Number)) report
    "ERROR : (tRCD_CHECK) : Active to column Access delay time violation."
    severity error;
    b0_last_column_access <= now after 1 ns;
  elsif (BA = "01" and ((ReadFlag = TRUE) or (WriteFlag = TRUE))) then
    assert ((now - b1_last_activate) >= tRCD (Part_Number)) report
    "ERROR : (tRCD_CHECK) : Active to column Access delay time violation."
    severity error;
    b1_last_column_access <= now after 1 ns;
  elsif (BA = "10" and ((ReadFlag = TRUE) or (WriteFlag = TRUE))) then
    assert ((now - b1_last_activate) >= tRCD (Part_Number)) report
    "ERROR : (tRCD_CHECK) : Active to column Access delay time violation."
    severity error;
    b2_last_column_access <= now after 1 ns;
  elsif (BA = "11" and ((ReadFlag = TRUE) or (WriteFlag = TRUE))) then
    assert ((now - b1_last_activate) >= tRCD (Part_Number)) report
    "ERROR : (tRCD_CHECK) : Active to column Access delay time violation."
    severity error;
    b3_last_column_access <= now after 1 ns;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tRCD_CHECK

----------------------------------------------------------------------------------
-----------------------tRAS_CHECK
tRAS_CHECK : process (BankActivatedFlag(0), BankActivatedFlag(1), BankActivatedFlag(2), BankActivatedFlag(3))
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BankActivatedFlag(0)'EVENT and BankActivatedFlag(0) = '0' and BankActivatedFlag(0)'LAST_VALUE = '1') then
    assert ((now - b0_last_activate) >= tRAS (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank0 active time minimum violation."
    severity error;
    assert ((now - b0_last_activate) <= tRASmax (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank0 active time maximum violation."
    severity error;
  end if;
  if (BankActivatedFlag(1)'EVENT and BankActivatedFlag(1) = '0' and BankActivatedFlag(1)'LAST_VALUE = '1') then
    assert ((now - b1_last_activate) >= tRAS (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank1 active time minimum violation."
    severity error;
    assert ((now - b1_last_activate) <= tRASmax (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank1 active time maximum violation."
    severity error;
  end if;
  if (BankActivatedFlag(2)'EVENT and BankActivatedFlag(2) = '0' and BankActivatedFlag(2)'LAST_VALUE = '1') then
    assert ((now - b2_last_activate) >= tRAS (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank2 active time minimum violation."
    severity error;
    assert ((now - b2_last_activate) <= tRASmax (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank2 active time maximum violation."
    severity error;
  end if;
  if (BankActivatedFlag(3)'EVENT and BankActivatedFlag(3) = '0' and BankActivatedFlag(3)'LAST_VALUE = '1') then
    assert ((now - b3_last_activate) >= tRAS (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank1 active time minimum violation."
    severity error;
    assert ((now - b3_last_activate) <= tRASmax (Part_Number)) report
    "ERROR : (tRAS_CHECK) : Bank1 active time maximum violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tRAS_CHECK

----------------------------------------------------------------------------------
------------------------tRP_CHECK
tRP_CHECK : process (BankActivateFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and (BankActivateFlag = TRUE)) then
    assert ((now - b0_last_precharge) >= tRP (Part_Number)) report
    "ERROR : (tRP_CHECK) : Precharge to active delay time violation."
    severity error;
  elsif (BA = "01" and (BankActivateFlag = TRUE)) then
    assert ((now - b1_last_precharge) >= tRP (Part_Number)) report
    "ERROR : (tRP_CHECK) : Precharge to active delay time violation."
    severity error;
  elsif (BA = "10" and (BankActivateFlag = TRUE)) then
    assert ((now - b2_last_precharge) >= tRP (Part_Number)) report
    "ERROR : (tRP_CHECK) : Precharge to active delay time violation."
    severity error;
  elsif (BA = "11" and (BankActivateFlag = TRUE)) then
    assert ((now - b3_last_precharge) >= tRP (Part_Number)) report
    "ERROR : (tRP_CHECK) : Precharge to active delay time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
--------------------tRP_CHECK

----------------------------------------------------------------------------------
-----------------------tRRD_CHECK
tRRD_CHECK : process (BankActivateFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b1_last_activate) >= tRRD (Part_Number)
        and (now - b2_last_activate) >= tRRD (Part_Number)
        and (now - b3_last_activate) >= tRRD (Part_Number)) report
    "ERROR : (tRRD_CHECK) : Active to the other bank active delay time violation."
    severity error;
  elsif (BA = "01" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b0_last_activate) >= tRRD (Part_Number)
        and (now - b2_last_activate) >= tRRD (Part_Number)
        and (now - b3_last_activate) >= tRRD (Part_Number)) report
    "ERROR : (tRRD_CHECK) : Active to the other bank active delay time violation."
    severity error;
  elsif (BA = "10" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b0_last_activate) >= tRRD (Part_Number)
        and (now - b1_last_activate) >= tRRD (Part_Number)
        and (now - b3_last_activate) >= tRRD (Part_Number)) report
    "ERROR : (tRRD_CHECK) : Active to the other bank active delay time violation."
    severity error;
  elsif (BA = "11" and (BankActivateFlag'EVENT and BankActivateFlag = TRUE)) then
    assert ((now - b0_last_activate) >= tRRD (Part_Number)
        and (now - b1_last_activate) >= tRRD (Part_Number)
        and (now - b2_last_activate) >= tRRD (Part_Number)) report
    "ERROR : (tRRD_CHECK) : Active to the other bank active delay time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tRRD_CHECK

----------------------------------------------------------------------------------
-----------------------tRRC_CHECK
tRRC_CHECK : process (AutoRefFlag)
begin
if (AutoRefFlag'EVENT and AutoRefFlag = TRUE and PUSCheckFinFlag = TRUE) then
  assert (now - last_aref_time >= tRRC (Part_Number)) report
  "ERROR : (tRRC_CHECK) : Auto Refresh cycle time violation!."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tRRC_CHECK

----------------------------------------------------------------------------------
-----------------------tCCD_CHECK
tCCD_CHECK : process (ReadFlag, WriteFlag, PrechargeFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (((ReadFlag'EVENT and ReadFlag = TRUE) or (WriteFlag'EVENT and WriteFlag = TRUE) or (PrechargeFlag'EVENT and PrechargeFlag = TRUE)) 
      and (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE)) then
  if (BA = "00") then
    assert ((now - b0_last_column_access) >= tCCD) report
    "ERROR : (tCCD_CHECK) : Column access to column access delay time violation."
    severity error;
  elsif (BA = "01") then
    assert ((now - b1_last_column_access) >= tCCD) report
    "ERROR : (tCCD_CHECK) : Column access to column access delay time violation."
    severity error;
  elsif (BA = "10") then
    assert ((now - b2_last_column_access) >= tCCD) report
    "ERROR : (tCCD_CHECK) : Column access to column access delay time violation."
    severity error;
  elsif (BA = "11") then
    assert ((now - b3_last_column_access) >= tCCD) report
    "ERROR : (tCCD_CHECK) : Column access to column access delay time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tCCD_CHECK

----------------------------------------------------------------------------------
-----------------------tDPL_CHECK
tDPL_CHECK : process (PrechargeFlag, PrechargeAllFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and (PrechargeFlag'EVENT and PrechargeFlag = TRUE)) then
    assert ((now - b0_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
  elsif (BA = "01" and (PrechargeFlag'EVENT and PrechargeFlag = TRUE)) then
    assert ((now - b1_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
  elsif (BA = "10" and (PrechargeFlag'EVENT and PrechargeFlag = TRUE)) then
    assert ((now - b2_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
  elsif (BA = "11" and (PrechargeFlag'EVENT and PrechargeFlag = TRUE)) then
    assert ((now - b3_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
  elsif (PrechargeAllFlag'EVENT and PrechargeAllFlag = TRUE) then
    assert ((now - b0_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
    assert ((now - b1_last_data_in) >= tDPL) report
    "ERROR : (tDPL_CHECK) : Last data in to precharge command delay violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tDPL_CHECK

----------------------------------------------------------------------------------
-----------------------tDAL_CHECK
tDAL_CHECK : process (BankActivateFlag)
  variable BA : std_logic_vector((NUM_OF_BANK_ADD - 1) downto 0) := "00";
begin
BA := (BA1, BA0);
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (BA = "00" and (BankActivateFlag = TRUE)) then
    assert ((now - b0_last_data_in) >= tDAL) report
    "ERROR : (tDAL_CHECK) : Last data in to active command delay violation."
    severity error;
  elsif (BA = "01" and (BankActivateFlag = TRUE)) then
    assert ((now - b1_last_data_in) >= tDAL) report
    "ERROR : (tDAL_CHECK) : Last data in to active command delay violation."
    severity error;
  elsif (BA = "10" and (BankActivateFlag = TRUE)) then
    assert ((now - b2_last_data_in) >= tDAL) report
    "ERROR : (tDAL_CHECK) : Last data in to active command delay violation."
    severity error;
  elsif (BA = "11" and (BankActivateFlag = TRUE)) then
    assert ((now - b3_last_data_in) >= tDAL) report
    "ERROR : (tDAL_CHECK) : Last data in to active command delay violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tDAL_CHECK

----------------------------------------------------------------------------------
-----------------------tMRD_CHECK
tMRD_CHECK : process (BankActivateFlag, PrechargeFlag, PrechargeAllFlag)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  assert ((now - last_mrs_set) >= tMRD) report
  "ERROR : (tMRD_CHECK) : MRS to new command delay violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tMRD_CHECK

----------------------------------------------------------------------------------
-----------------------tPDE_CHECK
tPDE_CHECK : process (CLK_INT)
begin
if (CLKStableFlag = TRUE and ModeRegisterSetFlag = TRUE) then
  if (PcgPwrDnExtFlag = TRUE or ActPwrDnExtFlag = TRUE) then
    assert ((now - last_pd_ext) >= tPDE) report
    "ERROR : (tPDE_CHECK) : Power down exit time violation."
    severity error;
    assert ((now - last_pd_ext) >= tPDE) report
    "ERROR : (tPDE_CHECK) : Power down exit time violation."
    severity error;
  end if;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tPDE_CHECK

----------------------------------------------------------------------------------
-----------------------tSRE_CHECK
tSRE_CHECK : process (CLK_INT)
begin
if (CLKStableFlag = TRUE and SelfRefExtFlag = TRUE and ModeRegisterSetFlag = TRUE) 
then
  assert ((now - last_sr_ext) >= tSRE) report
  "ERROR : (tSRE_CHECK) : Self refresh exit time violation."
  severity error;
end if;
end 
process;--------------------------------------------------------------------------
-------------------tSRE_CHECK

----------------------------------------------------------------------------------
---------------------------------------
End Behavioral_Model;
